# -*- coding: utf-8 -*-
"""
deserializer handlers base module.
"""

from abc import abstractmethod

from pyrin.converters.deserializer.exceptions import InvalidDeserializerTypeError
from pyrin.converters.deserializer.interface import AbstractDeserializerBase
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.core.globals import NULL
from pyrin.utils.string import remove_line_break_escapes


class DeserializerBase(AbstractDeserializerBase):
    """
    deserializer base class.
    """

    def __init__(self, **options):
        """
        initializes an instance of DeserializerBase.

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__()
        self._next_handler = None

        internal = options.get('internal')
        if internal is None:
            internal = False

        self._internal = internal

    def deserialize(self, value, **options):
        """
        deserializes the given value.

        returns `NULL` object if deserialization fails.

        :param object value: value to be deserialized.

        :keyword bool include_internal: specifies that any chained internal deserializer
                                        must also be used for deserialization. if set to
                                        False, only non-internal deserializers will be used.
                                        defaults to True if not provided.

        :returns: deserialized value.
        """

        deserialized_value = NULL
        include_internal = options.get('include_internal', True)
        if include_internal is not False or self.internal is False:
            deserialized_value = self._deserialize_operation(value, **options)

        if deserialized_value is NULL:
            if self._next_handler is not None:
                return self._next_handler.deserialize(value, **options)

        if isinstance(deserialized_value, str):
            deserialized_value = remove_line_break_escapes(deserialized_value)

        return deserialized_value

    def _deserialize_operation(self, value, **options):
        """
        deserializes the given value.

        it checks if the value is deserializable, if so, deserializes it.
        otherwise returns `NULL` object.
        this method could be overridden in other base subclasses if required.

        :param object value: value to be deserialized.

        :returns: deserialized value.
        """

        deserialized_value = NULL
        if self.is_deserializable(value, **options) is True:
            deserialized_value = self._deserialize(value, **options)

        return deserialized_value

    @abstractmethod
    def _deserialize(self, value, **options):
        """
        deserializes the given value.

        returns `NULL` object if deserialization fails.
        this method is intended to be overridden in subclasses.

        :param object value: value to be deserialized.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: deserialized value.
        """

        raise CoreNotImplementedError()

    def set_next(self, deserializer):
        """
        sets the next deserializer handler and returns it.

        :param DeserializerBase deserializer: deserializer instance to
                                              be set as next handler.

        :rtype: DeserializerBase
        """

        if deserializer is not None and not isinstance(deserializer, DeserializerBase):
            raise InvalidDeserializerTypeError('Input parameter [{instance}] is not '
                                               'an instance of [{base}].'
                                               .format(instance=deserializer,
                                                       base=DeserializerBase))

        self._next_handler = deserializer
        return deserializer

    def is_deserializable(self, value, **options):
        """
        gets a value indicating that the given input is deserializable.

        :param object value: value to be deserialized.

        :rtype: bool
        """

        return isinstance(value, self.accepted_type)

    @property
    def internal(self):
        """
        gets a value indicating that this deserializer is internal.

        internal deserializers will not be used for deserializing client inputs.

        :rtype: bool
        """

        return self._internal


class StringDeserializerBase(DeserializerBase):
    """
    string deserializer base class.
    """

    # these values will be used for accepted
    # formats that have no length restriction.
    UNDEF_LENGTH = None
    DEFAULT_MIN = 1
    DEFAULT_MAX = 1000000

    def __init__(self, **options):
        """
        initializes an instance of StringDeserializerBase.

        :keyword list[tuple[str, int, int]] accepted_formats: custom formats and their
                                                              min and max length that
                                                              this deserializer can
                                                              deserialize value from.

        :note accepted_formats: list[tuple[str format, int min_length, int max_length]]

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__(**options)

        self._accepted_formats = self.default_formats
        custom_accepted_formats = options.get('accepted_formats', [])
        self._accepted_formats.extend(custom_accepted_formats)

        # min and max accepted length of strings
        # to be deserialized by this deserializer.
        self._min_length, self._max_length = self._calculate_accepted_length()

    def is_deserializable(self, value, **options):
        """
        gets a value indicating that the given input is deserializable.

        :param object value: value to be deserialized.

        :rtype: bool
        """

        if super().is_deserializable(value, **options) \
                and self.is_valid_length(value.strip()):
            return True

        return False

    @property
    def accepted_type(self):
        """
        gets the accepted type for this deserializer.

        which could deserialize values from this type.

        :rtype: type[str]
        """

        return str

    @property
    def accepted_length(self):
        """
        gets the min and max accepted length of strings to be deserialized.

        :returns: tuple[int min, int max]
        :rtype: tuple[int, int]
        """

        return self._min_length, self._max_length

    @property
    def accepted_formats(self):
        """
        gets the accepted string formats that this deserializer can deserialize value from.

        :returns: list[tuple[str format, int min_length, int max_length]]
        :rtype: list[tuple[str, int, int]]
        """

        return self._accepted_formats

    def is_valid_length(self, value):
        """
        gets a value indicating that input value has valid length to be deserialized.

        :param str value: value to be deserialized.

        :rtype: bool
        """

        length = len(value.strip())
        min_length, max_length = self.accepted_length

        if length < min_length or length > max_length:
            return False

        return True

    def _calculate_accepted_length(self):
        """
        calculates the min and max accepted length of values to be deserialized.

        :returns: tuple[int min, int max]
        :rtype: tuple[int, int]
        """

        # if there is any format with length=UNDEF_LENGTH,
        # we should not enforce length restriction on values.
        min_length = None
        max_length = None
        if self.UNDEF_LENGTH in [item[1] for item in self.accepted_formats]:
            min_length = self.DEFAULT_MIN

        if self.UNDEF_LENGTH in [item[2] for item in self.accepted_formats]:
            max_length = self.DEFAULT_MAX

        if min_length is None:
            min_length = min([item[1] for item in self.accepted_formats])

        if max_length is None:
            max_length = max([item[2] for item in self.accepted_formats])

        return min_length, max_length

    @property
    @abstractmethod
    def default_formats(self):
        """
        gets default accepted formats that this deserializer could deserialize value from.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: list[tuple[str format, int min_length, int max_length]]
        :rtype: list[tuple[str, int, int]]
        """

        raise CoreNotImplementedError()


class StringPatternDeserializerBase(StringDeserializerBase):
    """
    string pattern deserializer base class.

    this class uses regex to determine whether a value is deserializable or not.
    """

    def __init__(self, **options):
        """
        initializes an instance of StringPatternDeserializerBase.

        :keyword list[tuple[Pattern, int, int]] accepted_formats: custom patterns and their
                                                                  min and max length that
                                                                  this deserializer can
                                                                  deserialize value from.

        :note accepted_formats: list[tuple[Pattern format, int min_length, int max_length]]

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__(**options)

    def is_deserializable(self, value, **options):
        """
        gets a value indicating that the given input is deserializable.

        if the value is deserializable, the matching `Pattern` would also be returned.
        otherwise None would be returned instead of `Pattern`.

        :param object value: value to be deserialized.

        :rtype: tuple[bool, Pattern]
        """

        if super().is_deserializable(value, **options):
            pattern = self.get_matching_pattern(value.strip())
            if pattern is not None:
                return True, pattern

        return False, None

    def _deserialize_operation(self, value, **options):
        """
        deserializes the given value.

        it checks if the value is deserializable, if so, deserializes it.
        otherwise returns `NULL` object.
        if the value is deserializable, the matching `Pattern` would also
        be injected into options with `matching_pattern` key to be used in
        `_deserialize()` method implementation.
        it's better to pop `matching_pattern` in `_deserialize()` method
        for usage instead of get.

        :param object value: value to be deserialized.

        :returns: deserialized value.
        """

        deserialized_value = NULL
        deserializable, pattern = self.is_deserializable(value, **options)
        if deserializable is True:
            options.update(matching_pattern=pattern)
            deserialized_value = self._deserialize(value.strip(), **options)

        return deserialized_value

    def get_matching_pattern(self, value):
        """
        gets the matching pattern for given value. returns None if no pattern found.

        :param object value: value to be deserialized.

        :rtype: Pattern
        """

        for pattern, min_length, max_length in self.accepted_formats:
            if pattern.match(value):
                return pattern

        return None



# -*- coding: utf-8 -*-
"""
orm scoping base module.
"""

from sqlalchemy.orm import scoped_session

from pyrin.database.orm.scoping.interface import AbstractScopedRegistryBase
from pyrin.database.orm.scoping.registry.request_scoped import RequestScopedRegistry
from pyrin.database.orm.scoping.registry.thread_scoped import ThreadScopedRegistry
from pyrin.database.orm.scoping.exceptions import InvalidRequestScopedRegistryTypeError, \
    InvalidThreadScopedRegistryTypeError, ScopedSessionIsAlreadyPresentError


class CoreScopedSession(scoped_session):
    """
    core scoped session class.

    it also supports atomic sessions.
    """

    def __init__(self, session_factory, scopefunc=None,
                 request_registry=RequestScopedRegistry,
                 thread_registry=ThreadScopedRegistry):
        """
        initializes an instance of CoreScopedSession.

        :param Session session_factory: a factory to create new `Session`
                                        instances. this is usually, but not
                                        necessarily, an instance of `sessionmaker`.

        :param function scopefunc: optional function which defines the current scope.
                                   if not passed, the `scoped_session` class object
                                   assumes `thread-local` scope, and will use a python
                                   `threading.local()` in order to maintain the current
                                   `Session`. if passed, the function should return a
                                   hashable token. this token will be used as the key
                                   in a dictionary in order to store and retrieve the
                                   current `Session`.

        :param type[AbstractScopedRegistryBase] request_registry: registry type to be used
                                                                  for request scoped sessions.
                                                                  defaults to
                                                                  `RequestScopedRegistry`
                                                                  if not provided.

        :param type[AbstractScopedRegistryBase] thread_registry: registry type to be used
                                                                 for thread scoped sessions.
                                                                 defaults to
                                                                 `ThreadScopedRegistry`
                                                                 if not provided.

        :raises InvalidRequestScopedRegistryTypeError: invalid request scoped
                                                       registry type error.

        :raises InvalidThreadScopedRegistryTypeError: invalid thread scoped
                                                      registry type error.
        """

        self.session_factory = session_factory

        message = 'Input parameter [{type_}] is not a subclass of [{base}].'
        if request_registry is None or not issubclass(request_registry,
                                                      AbstractScopedRegistryBase):

            raise InvalidRequestScopedRegistryTypeError(message
                                                        .format(type_=request_registry,
                                                                base=AbstractScopedRegistryBase))

        if thread_registry is None or not issubclass(thread_registry,
                                                     AbstractScopedRegistryBase):

            raise InvalidThreadScopedRegistryTypeError(message
                                                       .format(type_=thread_registry,
                                                               base=AbstractScopedRegistryBase))

        if scopefunc:
            self.registry = request_registry(session_factory, scopefunc)
        else:
            self.registry = thread_registry(session_factory)

    def __call__(self, atomic=False, **kw):
        """
        gets the current `Session` object.

        creating it using the `CoreScopedSession.session_factory` if not present.
        note that if there is an atomic object available, and `atomic=False` is
        provided, it always returns the available atomic object, but if `atomic=True`
        is provided it always creates a new atomic object.

        :param bool atomic: specifies that it must get a new atomic session.
                            defaults to False if not provided.

        :keyword object **kw: keyword arguments will be passed to the
                              `CoreScopedSession.session_factory` callable,
                              if an existing `Session` is not present. if
                              the `Session` is present and keyword arguments
                              have been passed, `ScopedSessionIsAlreadyPresentError`
                              is raised.

        :raises ScopedSessionIsAlreadyPresentError: scoped session is already present error.

        :rtype: CoreSession
        """

        if kw:
            if atomic is False and self.registry.has() is True:
                raise ScopedSessionIsAlreadyPresentError('Scoped session is already present, '
                                                         'no new arguments may be specified.')
            else:
                kw.update(atomic=atomic)
                session = self.session_factory(**kw)
                self.registry.set(session, atomic)
                return session
        else:
            return self.registry(atomic)

    def remove(self, atomic=False):
        """
        disposes the current `Session` objects of current scope, if present.

        this will first call `Session.close` method on the current `Session`,
        which releases any existing transactional/connection resources still
        being held. transactions specifically are rolled back. the `Session`
        is then discarded. upon next usage within the same scope, the
        `CoreScopedSession` will produce a new `Session` object.

        :param bool atomic: specifies that it must only dispose the current atomic
                            session of current scope. otherwise, it disposes all
                            sessions of current scope.
                            defaults to False if not provided.
        """

        self._remove_atomic(atomic=atomic)
        if atomic is False:
            self._remove()

    def _remove(self):
        """
        disposes the current normal `Session` object of current scope, if present.

        this will first call `Session.close` method on the current `Session`,
        which releases any existing transactional/connection resources still
        being held. transactions specifically are rolled back. the `Session`
        is then discarded. upon next usage within the same scope, the
        `CoreScopedSession` will produce a new `Session` object.
        """

        session = self.registry.get()
        if session is not None:
            session.close()
            self.registry.clear()

    def _remove_atomic(self, atomic=False):
        """
        disposes the current atomic `Session` objects of current scope, if present.

        this will first call `Session.close` method on the current `Session`,
        which releases any existing transactional/connection resources still
        being held. transactions specifically are rolled back. the `Session`
        is then discarded. upon next usage within the same scope, the
        `CoreScopedSession` will produce a new atomic `Session` object.

        :param bool atomic: specifies that it must only dispose the current atomic
                            session of current scope. otherwise, it disposes all
                            atomic sessions of current scope.
                            defaults to False if not provided.
        """

        if atomic is False:
            self._remove_all_atomic()
        else:
            atomic_session = self.registry.get(atomic=True)
            if atomic_session is not None:
                atomic_session.close()
                self.registry.clear(atomic=True)

    def _remove_all_atomic(self):
        """
        disposes all atomic `Session` objects of current scope, if present.

        this will first call `Session.close` method on the current `Session`,
        which releases any existing transactional/connection resources still
        being held. transactions specifically are rolled back. the `Session`
        is then discarded. upon next usage within the same scope, the
        `CoreScopedSession` will produce a new atomic `Session` object.
        """

        atomic_sessions = self.registry.get_all_atomic()
        for session in atomic_sessions:
            session.close()

        self.registry.clear_all_atomic()

    def commit_all(self, atomic=False):
        """
        commits all sessions related to current scope.

        :param bool atomic: specifies that it must only commit the current atomic session
                            of current scope. otherwise, it commits all sessions of current
                            scope. defaults to False if not provided.
        """

        self._commit_atomic(atomic=atomic)
        if atomic is False:
            self._commit()

    def _commit(self):
        """
        commits the normal session of current scope, if available.

        note that current scope could have at most one normal
        session at the same time.
        """

        session = self.registry.get()
        if session is not None:
            session.commit()

    def _commit_atomic(self, atomic=False):
        """
        commits current atomic session of current scope, if available.

        :param bool atomic: specifies that it must only commit the current atomic session
                            of current scope. otherwise, it commits all atomic sessions
                            of current scope. defaults to False if not provided.
        """

        if atomic is False:
            self._commit_all_atomic()
        else:
            atomic_session = self.registry.get(atomic=True)
            if atomic_session is not None:
                atomic_session.commit()

    def _commit_all_atomic(self):
        """
        commits all atomic sessions of current scope, if available.
        """

        atomic_sessions = self.registry.get_all_atomic()
        for session in atomic_sessions:
            session.commit()

    def rollback_all(self, atomic=False):
        """
        rollbacks all sessions related to current scope.

        :param bool atomic: specifies that it must only rollback the atomic session of current
                            scope. otherwise, it commits all sessions of current scope.
                            defaults to False if not provided.
        """

        self._rollback_atomic(atomic=atomic)
        if atomic is False:
            self._rollback()

    def _rollback(self):
        """
        rollbacks the normal session of current scope, if available.

        note that current scope could have at most one normal
        session at the same time.
        """

        session = self.registry.get()
        if session is not None:
            session.rollback()

    def _rollback_atomic(self, atomic=False):
        """
        rollbacks the atomic session of current scope, if available.

        :param bool atomic: specifies that it must only rollback the current atomic
                            session of current scope. otherwise, it rollbacks all atomic
                            sessions of current scope. defaults to False if not provided.
        """

        if atomic is False:
            self._rollback_all_atomic()
        else:
            atomic_session = self.registry.get(atomic=True)
            if atomic_session is not None:
                atomic_session.rollback()

    def _rollback_all_atomic(self):
        """
        rollbacks all atomic sessions of current scope, if available.
        """

        atomic_sessions = self.registry.get_all_atomic()
        for session in atomic_sessions:
            session.rollback()



# -*- coding: utf-8 -*-
"""
deserializer handlers timedelta module.
"""

import re

from datetime import timedelta

import pyrin.converters.deserializer.services as deserializer_services

from pyrin.core.globals import NULL
from pyrin.converters.deserializer.decorators import deserializer
from pyrin.converters.deserializer.handlers.base import StringPatternDeserializerBase


@deserializer()
class TimedeltaDeserializer(StringPatternDeserializerBase):
    """
    timedelta deserializer class.
    """

    # matches the timedelta definition inside string.
    # example: timedelta(23, 1, 1)
    # it does not support keyword arguments, only positional arguments are supported.
    # matching is case-sensitive.
    TIMEDELTA_REGEX = re.compile(r'^timedelta\([\d, .]*\)$')

    def __init__(self, **options):
        """
        creates an instance of TimedeltaDeserializer.

        :keyword list[tuple[Pattern, int, int]] accepted_formats: a list of custom accepted
                                                                  formats and their min and
                                                                  max length for timedelta
                                                                  deserialization.

        :note accepted_formats: list[tuple[Pattern format, int min_length, int max_length]]
        """

        options.update(internal=True)
        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.

        returns `NULL` object if deserialization fails.
        but if any error occurs, it will be raised.

        :param str value: value to be deserialized.

        :keyword Pattern matching_pattern: the pattern that has matched the value.

        :rtype: timedelta
        """

        args = value.replace('timedelta', '')
        args = deserializer_services.deserialize(args)
        if isinstance(args, tuple):
            return timedelta(*args)

        return NULL

    @property
    def default_formats(self):
        """
        gets default accepted formats that this deserializer could deserialize value from.

        :returns: list[tuple[Pattern format, int min_length, int max_length]]
        :rtype: list[tuple[Pattern, int, int]]
        """

        return [(self.TIMEDELTA_REGEX, 11, self.UNDEF_LENGTH)]



# -*- coding: utf-8 -*-
"""
validator component module.
"""

from pyrin.application.decorators import component
from pyrin.validator import ValidatorPackage
from pyrin.validator.manager import ValidatorManager
from pyrin.application.structs import Component


@component(ValidatorPackage.COMPONENT_NAME)
class ValidatorComponent(Component, ValidatorManager):
    """
    validator component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
caching local containers dict module.
"""

from collections import OrderedDict

from pyrin.caching.local.containers.base import LocalCacheContainerBase


class OrderedDictContainer(OrderedDict, LocalCacheContainerBase):
    """
    ordered dict container class.

    it is actually a regular ordered dict.
    it is inherited from `OrderedDict` instead of `dict` to let efficiently
    remove items in `FIFO` or `LIFO` order when items count limit is reached.
    """

    def set(self, key, value, *args, **kwargs):
        """
        sets a new value into cached items.

        :param object key: hashable key of the cache to be registered.
        :param object value: value to be cached.
        """

        self[key] = value

    def slice(self, count, last=False):
        """
        gets a slice of items of this container with the count length.

        :param int count: number of items to include in slice.

        :param bool last: specifies that slice must include items from end.
                          defaults to False if not provided.

        :returns: iterable
        """

        items = list(self.keys())
        if last is not True:
            return items[:count]
        else:
            return items[count * -1:]



# -*- coding: utf-8 -*-
"""
database services module.
"""

from pyrin.application.services import get_component
from pyrin.database import DatabasePackage


def get_current_store(**kwargs):
    """
    gets current database store.

    note that this method will always get the correct session based on available
    context. so if you are in an atomic context, it gets you the correct atomic
    session, but if you are not in an atomic context, it will get you the related
    session to current scope.

    :keyword object **kwargs: keyword arguments will be passed to the
                              `CoreScopedSession.session_factory` callable
                              to configure the new session that's being
                              created, if an existing session is not present.
                              if the session is present and keyword arguments
                              have been passed, an error will be raised.

    :raises ScopedSessionIsAlreadyPresentError: scoped session is already present error.

    :returns: database session
    :rtype: CoreSession
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_current_store(**kwargs)


def get_current_session_factory():
    """
    gets current database session factory.

    this method should not be used directly for data manipulation.
    use `get_current_store` method instead.

    :returns: database session factory
    :rtype: CoreScopedSession
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_current_session_factory()


def get_atomic_store(**kwargs):
    """
    gets an atomic database store.

    atomic is meaning a new session with a new transaction. note that it's normally
    not needed to get an atomic session manually, instead you could use `@atomic`
    decorator to provide you an atomic session. but if you really need to get an atomic
    session manually, you have to manually remove that session from corresponding
    session factory after you've done. otherwise, unexpected behaviour may occur.
    so if you get an atomic session, and don't remove it, after that if you get another
    session in the same scope, you will get the same exact atomic session. but if you
    get an atomic session, and remove it from corresponding session factory after you've
    done, after that if you get a session, it will get you a session related to current
    scope. this is why it's recommended not to get an atomic session manually, and
    instead use `@atomic` decorator when you need an atomic session.

    :keyword bool expire_on_commit: expire atomic session after commit.
                                    it is useful to set it to True if
                                    the atomic function does not return
                                    any entities for post-processing.
                                    defaults to False if not provided.

    :keyword object **kwargs: keyword arguments will be passed to the
                              `CoreScopedSession.session_factory` callable
                              to configure the new atomic session that's
                              being created.

    :returns: database session
    :rtype: CoreSession
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_atomic_store(**kwargs)


def finalize_transaction(response, **options):
    """
    this method will finalize database transaction of each request.

    this method will finalize both normal and atomic sessions of
    current request if available.
    we should not raise any exception in finalize transaction hook,
    so we return an error response in case of any exception.
    note that normally you should never call this method manually.

    :param CoreResponse response: response object.

    :rtype: CoreResponse | tuple
    """

    return get_component(DatabasePackage.COMPONENT_NAME).finalize_transaction(response,
                                                                              **options)


def cleanup_session(exception):
    """
    this method will cleanup database session of each request.

    this method will cleanup both normal and atomic sessions of
    current request if available.
    in case of any unhandled exception. we should not raise any exception
    in teardown request handlers, so we just log the exception.
    note that normally you should never call this method manually.

    :param Exception exception: exception instance.
    """

    return get_component(DatabasePackage.COMPONENT_NAME).cleanup_session(exception)


def register_session_factory(instance, **options):
    """
    registers a new session factory or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding an instance which it's `request_bounded` is already available
    in registered session factories.

    :param AbstractSessionFactoryBase instance: session factory to be registered.
                                                it must be an instance of
                                                AbstractSessionFactoryBase.

    :keyword bool replace: specifies that if there is another registered
                           session factory with the same `request_bounded`,
                           replace it with the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidSessionFactoryTypeError: invalid session factory type error.
    :raises DuplicatedSessionFactoryError: duplicated session factory error.
    """

    return get_component(DatabasePackage.COMPONENT_NAME).register_session_factory(instance,
                                                                                  **options)


def register_bind(entity, bind_name, **options):
    """
    binds the given model class with specified bind database.

    :param type[BaseEntity] entity: base entity subclass to be bounded.
    :param str bind_name: bind name to be associated with the model class.

    :raises InvalidEntityTypeError: invalid entity type error.
    """

    return get_component(DatabasePackage.COMPONENT_NAME).register_bind(entity, bind_name,
                                                                       **options)


def configure_session_factories():
    """
    configures all application session factories.

    normally, you should not call this method manually.

    :raises InvalidDatabaseBindError: invalid database bind error.
    """

    return get_component(DatabasePackage.COMPONENT_NAME).configure_session_factories()


def get_default_engine():
    """
    gets database default engine.

    :rtype: Engine
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_default_engine()


def get_bounded_engines():
    """
    gets database bounded engines.

    :returns: dict[str bind_name: Engine engine]
    :rtype: dict
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_bounded_engines()


def get_entity_to_engine_map():
    """
    gets entity to engine map.

    :returns: dict[type entity, Engine engine]
    :rtype: dict
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_entity_to_engine_map()


def get_table_name_to_engine_map():
    """
    gets table name to engine map.

    :returns: dict[str table_name, Engine engine]
    :rtype: dict
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_table_name_to_engine_map()


def register_hook(instance):
    """
    registers the given instance into database hooks.

    :param DatabaseHookBase instance: database hook instance to be registered.

    :raises InvalidDatabaseHookTypeError: invalid database hook type error.
    """

    return get_component(DatabasePackage.COMPONENT_NAME).register_hook(instance)


def get_table_engine(table_name):
    """
    gets the database engine which the provided table name is bounded to.

    :param str table_name: table name to get its bounded engine.

    :rtype: Engine
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_table_engine(table_name)


def get_bind_name_engine(bind_name):
    """
    gets the database engine which the provided bind name is bounded to.

    :param str bind_name: bind name to get its bounded engine.

    :raises InvalidDatabaseBindError: invalid database bind error.

    :rtype: Engine
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_bind_name_engine(bind_name)


def get_entity_engine(entity):
    """
    gets the database engine which the provided entity class is bounded to.

    :param BaseEntity entity: entity class to get its bounded engine.

    :rtype: Engine
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_entity_engine(entity)


def get_bind_config_section_name(bind_name):
    """
    gets the bind config section name for given bind name and currently
    active environment in 'database.ini' from 'database.binds.ini' file.

    :param str bind_name: bind name to get its section name.

    :rtype: str
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_bind_config_section_name(bind_name)


def get_default_database_name():
    """
    gets default database name.

    :rtype: str
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_default_database_name()


def get_configs_prefix():
    """
    gets the configs prefix for sqlalchemy keys in database config store.

    it gets the value of `configs_prefix` key from database config store.

    :rtype: str
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_configs_prefix()


def get_all_database_names():
    """
    gets all database names defined in application.

    it returns all available database names, even those
    that do not have any entity bounded to them.
    the result also includes the default database name.

    :rtype: list[str]
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_all_database_names()


def get_database_bind_names():
    """
    gets all database bind names defined in application.

    it returns all available database bind names, even those
    that do not have any entity bounded to them.

    :rtype: list[str]
    """

    return get_component(DatabasePackage.COMPONENT_NAME).get_database_bind_names()



# -*- coding: utf-8 -*-
"""
string normalizer handlers filter module.
"""

from pyrin.utilities.string.normalizer.decorators import string_normalizer
from pyrin.utilities.string.normalizer.enumerations import NormalizerEnum
from pyrin.utilities.string.normalizer.handlers.base import ReplaceNormalizerBase, \
    FilterNormalizerBase


@string_normalizer()
class FilterNormalizer(FilterNormalizerBase):
    """
    filter normalizer class.

    this normalizer will filter provided values from string.
    """

    def __init__(self, **options):
        """
        initializes an instance of FilterNormalizer.

        :keyword list[str] filters: values to be removed from string.

        :keyword dict filter_map: a dict of keys and values to be used
                                  for replacement.

        :note filters, filter_map: the `filters` value have precedence
                                   over `filter_map` value if they have
                                   common keys.

        :raises FiltersMustBeListError: filters must be list error.
        :raises FilterMapMustBeDictError: filter map must be dict error.
        """

        super().__init__(NormalizerEnum.FILTER, 430, **options)


@string_normalizer()
class PersianSignFilterNormalizer(ReplaceNormalizerBase):
    """
    persian sign filter normalizer class.

    this normalizer removes all common persian signs from string.
    """

    def __init__(self, **options):
        """
        initializes an instance of PersianSignFilterNormalizer.
        """

        replace_map = {'،': None, '؛': None, 'ـ': None,
                       '٬': None, '٪': None, '؟': None}

        super().__init__(NormalizerEnum.PERSIAN_SIGN, 420, replace_map, **options)


@string_normalizer()
class LatinSignFilterNormalizer(ReplaceNormalizerBase):
    """
    latin sign filter normalizer class.

    this normalizer removes all common latin signs from string.
    """

    def __init__(self, **options):
        """
        initializes an instance of LatinSignFilterNormalizer.
        """

        replace_map = {',': None, '.': None, ':': None, ';': None, '-': None,
                       '+': None, '_': None, '`': None, '"': None, "'": None,
                       '@': None, '#': None, '%': None, '^': None, '*': None,
                       '$': None, '(': None, ')': None, '{': None, '}': None,
                       '|': None, '~': None, '?': None, '!': None, '/': None,
                       '\\': None, '<': None, '>': None, '[': None, ']': None}

        super().__init__(NormalizerEnum.LATIN_SIGN, 410, replace_map, **options)



# -*- coding: utf-8 -*-
"""
model declarative module.

important note:

each `pyrin` package that references to this module, must
add `pyrin.database.model` into its `DEPENDS` list.
"""

from sqlalchemy.orm import DeclarativeMeta

import pyrin.database.model.services as model_services

from pyrin.database.model.base import BaseEntity


class CoreEntity(BaseEntity, metaclass=DeclarativeMeta):
    """
    core entity class.

    it should be used as the base class for all application concrete models.

    it is also possible to implement a new customized declarative base class for your
    application models. to do this you must subclass it from `BaseEntity`, because
    application will check isinstance() on `BaseEntity` type to detect models. then
    implement customized or new features in your subclassed `BaseEntity`. keep in mind
    that you should not subclass directly from `CoreEntity`, because it is a `DeclarativeMeta`
    and sqlalchemy raises an error if you subclass from this as your new declarative
    base class. note that your application must have a unique declarative base class for all
    models, do not mix the use of your new base class and `CoreEntity`, otherwise you will
    face problems in migrations and also multi-database environments.
    """

    __abstract__ = True
    registry = model_services.get_mapper_registry()
    metadata = model_services.get_mapper_registry().metadata



# -*- coding: utf-8 -*-
"""
serializer interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.structs import CoreObject, MultiSingletonMeta
from pyrin.core.exceptions import CoreNotImplementedError


class SerializerSingletonMeta(MultiSingletonMeta):
    """
    serializer singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractSerializerBase(CoreObject, metaclass=SerializerSingletonMeta):
    """
    abstract serializer base class.
    """

    @abstractmethod
    def serialize(self, value, **options):
        """
        serializes the given value.

        returns serialized value on success or `NULL` object if serialization fails.

        :param object value: value to be serialized.

        :keyword ResultSchema result_schema: result schema instance to be
                                             used to create computed columns.
                                             defaults to None if not provided.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: serialized object
        :rtype: dict | list[dict]
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def accepted_type(self):
        """
        gets the accepted type for this serializer.

        which could serialize values from this type.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: type
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
celery cli enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class CeleryCLIHandlersEnum(CoreEnum):
    """
    celery cli handlers enum.
    """

    AMQP = 'amqp'
    BEAT = 'beat'
    CALL = 'call'
    CONTROL = 'control'
    EVENTS = 'events'
    GRAPH = 'graph'
    HELP = 'help'
    LIST = 'list'
    INSPECT = 'inspect'
    LOGTOOL = 'logtool'
    MIGRATE = 'migrate'
    PURGE = 'purge'
    REPORT = 'report'
    RESULT = 'result'
    SHELL = 'shell'
    STATUS = 'status'
    UPGRADE = 'upgrade'
    WORKER = 'worker'



# -*- coding: utf-8 -*-
"""
database migration hooks module.
"""

import pyrin.database.migration.services as migration_services

from pyrin.application.decorators import application_hook
from pyrin.application.hooks import ApplicationHookBase
from pyrin.database.decorators import database_hook
from pyrin.database.hooks import DatabaseHookBase


@database_hook()
class DatabaseHook(DatabaseHookBase):
    """
    database hook class.
    """

    def after_session_factories_configured(self):
        """
        this method will be called after all database session factories have been configured.
        """

        migration_services.configure_migration_data()


@application_hook()
class ApplicationHook(ApplicationHookBase):
    """
    application hook class.
    """

    def after_application_loaded(self):
        """
        this method will be called after application has been loaded.
        """

        migration_services.rebuild_schema()



# -*- coding: utf-8 -*-
"""
orm sql schema mixin module.
"""

from sqlalchemy import BigInteger, Integer, Sequence

import pyrin.utils.misc as misc_utils
import pyrin.utils.unique_id as uuid_utils

from pyrin.database.orm.types.custom import GUID
from pyrin.database.orm.sql.schema.exceptions import SequenceColumnTypeIsInvalidError


class SequenceColumnMixin:
    """
    sequence column mixin class.
    """

    DEFAULT_CACHE = 100
    DEFAULT_TYPE = BigInteger

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of SequenceColumnMixin.

        :param str name: the name of this column as represented in the database.
                         this argument may be the first positional argument, or
                         specified via keyword.

        :param TypeEngine type_: the column's type, indicated using an instance which
                                 if no arguments are required for the type, the class of
                                 the type can be sent as well.
                                 this argument may be the second positional argument, or
                                 specified via keyword.

        :param object *args: additional positional arguments include various
                             `SchemaItem` derived constructs which will be applied
                             as options to the column.

        :keyword str sequence: sequence name to be generated for this column.
                               this value is required, but has to be set as
                               keyword to prevent errors.

        :keyword str doc: optional string that can be used by the ORM or similar
                          to document attributes on the python side.

        :keyword str key: an optional string identifier which will identify this
                          `Column` object on the `Table`.

        :keyword bool index: when `True`, indicates that the column is indexed.

        :keyword dict info: optional data dictionary which will be populated into the
                            `SchemaItem.info` attribute of this object.

        :keyword bool nullable: when set to `False`, will cause the `Not NULL`
                                phrase to be added when generating ddl for the column.
                                defaults to False if not provided.

        :keyword bool primary_key: if `True`, marks this column as a primary key
                                   column. multiple columns can have this flag set to
                                   specify composite primary keys.

        :keyword bool quote: force quoting of this column's name on or off,
                             corresponding to `True` or `False`. when left at its default
                             of `None`, the column identifier will be quoted according to
                             whether the name is case sensitive (identifiers with at least one
                             upper case character are treated as case sensitive), or if it's a
                             reserved word.

        :keyword bool unique: when `True`, indicates that this column contains a
                              unique constraint, or if `index` is `True` as well, indicates
                              that the `index` should be created with the unique flag.

        :keyword bool system: when `True`, indicates this is a system column,
                              that is a column which is automatically made available by the
                              database, and should not be included in the columns list for a
                              `create table` statement.

        :keyword str comment: optional string that will render an sql comment
                              on table creation.

        :keyword bool allow_read: specifies that the column should be
                                  included in entity to dict conversion.
                                  defaults to True if not provided.

        :keyword bool allow_write: specifies that the column should be
                                   populated on conversion from dict.
                                   defaults to False if not provided.

        :keyword bool validated: specifies that an automatic validator for this column
                                 must be registered, that is usable through validator
                                 services for create and update.
                                 defaults to False if not provided.

        :keyword bool validated_find: specifies that an automatic find validator for this
                                      column must be registered, that is usable through
                                      validator services for find. defaults to `validated`
                                      value if not provided.

        :keyword bool validated_range: specifies that automatic find range validators for this
                                       column must be registered, that is usable through
                                       validator services for find. defaults to `validated_find`
                                       value if not provided.
                                       note that find range validators are constructed with
                                       names `from_*` and `to_*` for given column if it
                                       is a number or any variant of date and time.
                                       if the type of column is anything else or it is a
                                       primary key, no range validators will be registered
                                       for it and this value will be ignored.

        :keyword int cache: cache size for sequence.
                            defaults to `DEFAULT_CACHE`, if not provided.
                            to disable cache, you can pass it as None or `0`.
                            note that cache is per session, so if you stop the
                            connection to database and start it again, the new
                            session will get its own cache. but it's good for
                            performance to have cache on sequences, gaps are not
                            bad at all.

        :raises SequenceColumnTypeIsInvalidError: sequence column type is invalid error.
        """

        args = list(args)
        name, type_ = self._extract_name_and_type(args, kwargs)
        if type_ is None:
            type_ = self.DEFAULT_TYPE

        if not misc_utils.is_subclass_or_instance(type_, Integer):
            raise SequenceColumnTypeIsInvalidError('The sequence column type must be an '
                                                   'instance or subclass of [{integer}].'
                                                   .format(integer=Integer))

        cache = kwargs.pop('cache', self.DEFAULT_CACHE)
        sequence = kwargs.pop('sequence', None)
        sequence_kwargs = dict()
        if cache is not None and cache > 0:
            sequence_kwargs.update(cache=cache)

        sequence_instance = Sequence(sequence, **sequence_kwargs)
        # this is to prevent sqlalchemy errors.
        if sequence is not None:
            args.append(sequence_instance)

        kwargs.setdefault('allow_write', False)
        kwargs.setdefault('nullable', False)

        kwargs.update(name=name, type_=type_,
                      autoincrement=False, min_value=1,
                      default=sequence_instance,
                      server_default=sequence_instance.next_value())

        kwargs.pop('onupdate', None)
        kwargs.pop('server_onupdate', None)

        super().__init__(*args, **kwargs)


class GUIDColumnMixin:
    """
    guid column mixin class.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of GUIDColumnMixin.

        :param str name: the name of this column as represented in the database.
                         this argument may be the first positional argument, or
                         specified via keyword.

        :param object *args: additional positional arguments include various
                             `SchemaItem` derived constructs which will be applied
                             as options to the column.

        :keyword str doc: optional string that can be used by the ORM or similar
                          to document attributes on the python side.

        :keyword str key: an optional string identifier which will identify this
                          `Column` object on the `Table`.

        :keyword bool index: when `True`, indicates that the column is indexed.

        :keyword dict info: optional data dictionary which will be populated into the
                            `SchemaItem.info` attribute of this object.

        :keyword bool nullable: when set to `False`, will cause the `Not NULL`
                                phrase to be added when generating ddl for the column.
                                defaults to False if not provided.

        :keyword bool primary_key: if `True`, marks this column as a primary key
                                   column. multiple columns can have this flag set to
                                   specify composite primary keys.

        :keyword bool quote: force quoting of this column's name on or off,
                             corresponding to `True` or `False`. when left at its default
                             of `None`, the column identifier will be quoted according to
                             whether the name is case sensitive (identifiers with at least one
                             upper case character are treated as case sensitive), or if it's a
                             reserved word.

        :keyword bool unique: when `True`, indicates that this column contains a
                              unique constraint, or if `index` is `True` as well, indicates
                              that the `index` should be created with the unique flag.

        :keyword bool system: when `True`, indicates this is a system column,
                              that is a column which is automatically made available by the
                              database, and should not be included in the columns list for a
                              `create table` statement.

        :keyword str comment: optional string that will render an sql comment
                              on table creation.

        :keyword bool allow_read: specifies that the column should be
                                  included in entity to dict conversion.
                                  defaults to True if not provided.

        :keyword bool allow_write: specifies that the column should be
                                   populated on conversion from dict.
                                   defaults to False if not provided.

        :keyword bool validated: specifies that an automatic validator for this column
                                 must be registered, that is usable through validator
                                 services for create and update.
                                 defaults to False if not provided.

        :keyword bool validated_find: specifies that an automatic find validator for this
                                      column must be registered, that is usable through
                                      validator services for find. defaults to `validated`
                                      value if not provided.
        """

        args = list(args)
        name, type_ = self._extract_name_and_type(args, kwargs)

        kwargs.setdefault('allow_write', False)
        kwargs.setdefault('nullable', False)

        kwargs.update(name=name, type_=GUID, autoincrement=False,
                      default=uuid_utils.generate_uuid4)

        kwargs.pop('server_default', None)
        kwargs.pop('onupdate', None)
        kwargs.pop('server_onupdate', None)

        super().__init__(*args, **kwargs)


class TypeMixin:
    """
    type mixin class.
    """

    _column_type = None

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of TypeMixin.

        :param str name: the name of this column as represented in the database.
                         this argument may be the first positional argument, or
                         specified via keyword.

        :param object *args: additional positional arguments include various
                             `SchemaItem` derived constructs which will be applied
                             as options to the column.

        :keyword bool autoincrement: set up `auto increment` semantics for an
                                     integer primary key column.

        :keyword callable | object default: a scalar, python callable or `ColumnElement`
                                            expression representing the default value
                                            for this column, which will be invoked upon
                                            insert if this column is otherwise not
                                            specified in the values clause of the insert.

        :keyword str doc: optional string that can be used by the ORM or similar
                          to document attributes on the python side.

        :keyword str key: an optional string identifier which will identify this
                          `Column` object on the `Table`.

        :keyword bool index: when `True`, indicates that the column is indexed.

        :keyword dict info: optional data dictionary which will be populated into the
                            `SchemaItem.info` attribute of this object.

        :keyword bool nullable: when set to `False`, will cause the `Not NULL`
                                phrase to be added when generating ddl for the column.

        :keyword callable | object onupdate: a scalar, python callable, or
                                             `ClauseElement` representing a default
                                             value to be applied to the column within update
                                             statements, which will be invoked upon update
                                             if this column is not present in the set
                                             clause of the update.

        :keyword bool primary_key: if `True`, marks this column as a primary key
                                   column. multiple columns can have this flag set to
                                   specify composite primary keys.

        :keyword str | ClauseElement | TextClause server_default: a `FetchedValue` instance,
                                                                  str, unicode or `text`
                                                                  construct representing the ddl
                                                                  default value for the column.

        :keyword str | ClauseElement | TextClause server_onupdate: a `FetchedValue` instance
                                                                   representing a database-side
                                                                   default generation function,
                                                                   such as a trigger. this
                                                                   indicates to sqlalchemy
                                                                   that a newly generated value
                                                                   will be available after
                                                                   updates. this construct
                                                                   does not actually implement
                                                                   any kind of generation
                                                                   function within the database,
                                                                   which instead must be specified
                                                                   separately.

        :keyword bool quote: force quoting of this column's name on or off,
                             corresponding to `True` or `False`. when left at its default
                             of `None`, the column identifier will be quoted according to
                             whether the name is case sensitive (identifiers with at least one
                             upper case character are treated as case sensitive), or if it's a
                             reserved word.

        :keyword bool unique: when `True`, indicates that this column contains a
                              unique constraint, or if `index` is `True` as well, indicates
                              that the `index` should be created with the unique flag.

        :keyword bool system: when `True`, indicates this is a system column,
                              that is a column which is automatically made available by the
                              database, and should not be included in the columns list for a
                              `create table` statement.

        :keyword str comment: optional string that will render an sql comment
                              on table creation.

        :keyword bool allow_read: specifies that the column should be
                                  included in entity to dict conversion.
                                  defaults to True if not provided.

        :keyword bool allow_write: specifies that the column should be
                                   populated on conversion from dict.
                                   defaults to True if not provided.

        :keyword object | callable min_value: minimum value that this column could have.
                                              it could also be a callable without any inputs.
                                              if a non-callable is provided and the column is
                                              not a primary key and also column name is provided,
                                              it will result in check constraint generation on
                                              database. otherwise it will be ignored and could
                                              be used in validators.
                                              defaults to None if not provided.

        :keyword object | callable max_value: maximum value that this column could have.
                                              it could also be a callable without any inputs.
                                              if a non-callable is provided and the column is
                                              not a primary key and also column name is provided,
                                              it will result in check constraint generation on
                                              database. otherwise it will be ignored and could
                                              be used in validators.
                                              defaults to None if not provided.

        :keyword list | callable check_in: list of valid values for this column.
                                           it could also be a callable without any inputs.
                                           if a non-callable is provided and the column is
                                           not a primary key and also column name is provided,
                                           it will result in check constraint generation on
                                           database. otherwise it will be ignored and could
                                           be used in validators.
                                           defaults to None if not provided.

        :keyword list | callable check_not_in: list of invalid values for this column.
                                               it could also be a callable without any inputs.
                                               if a non-callable is provided and the column is
                                               not a primary key and also column name is provided,
                                               it will result in check constraint generation on
                                               database. otherwise it will be ignored and could
                                               be used in validators.
                                               defaults to None if not provided.

        :note check_in, check_not_in: only one of these options could be provided.
                                      otherwise it raises an error.

        :keyword bool validated: specifies that an automatic validator for this column
                                 must be registered, that is usable through validator
                                 services for create and update.
                                 defaults to False if not provided.

        :keyword bool validated_find: specifies that an automatic find validator for this
                                      column must be registered, that is usable through
                                      validator services for find. defaults to `validated`
                                      value if not provided.

        :keyword bool validated_range: specifies that automatic find range validators for this
                                       column must be registered, that is usable through
                                       validator services for find. defaults to `validated_find`
                                       value if not provided.
                                       note that find range validators are constructed with
                                       names `from_*` and `to_*` for given column if it
                                       is a number or any variant of date and time.
                                       if the type of column is anything else or it is a
                                       primary key, no range validators will be registered
                                       for it and this value will be ignored.
        """

        args = list(args)
        name, type_ = self._extract_name_and_type(args, kwargs)
        kwargs.update(name=name, type_=self._column_type)

        super().__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
orm session base module.
"""

from sqlalchemy.orm import Session
from sqlalchemy.util import EMPTY_DICT
from sqlalchemy.sql.elements import TextClause

import pyrin.database.services as database_services
import pyrin.database.orm.sql.extractor.services as extractor_services

from pyrin.core.globals import _
from pyrin.core.structs import CoreObject
from pyrin.database.exceptions import InvalidDatabaseBindError
from pyrin.database.orm.session.exceptions import TransientSQLExpressionRequiredError


class CoreSession(Session, CoreObject):
    """
    core session class.

    all application sessions must be an instance this.
    """

    # these keywords are forbidden when 'execute' method is called
    # with 'transient=True' and a raw sql is given.
    NON_TRANSIENT_KEYWORDS = ('commit', 'flush', 'rollback', 'alter ', 'create ', 'drop ')

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of CoreSession.

        all positional and keyword arguments will be passed to underlying `Session`
        class. for details on inputs, see `Session` docstrings.

        :param object args: all positional arguments.

        :keyword object kwargs: all keyword arguments.

        :keyword bool atomic: specifies that this session is an atomic session.
                              defaults to False if not provided.
        """

        self._atomic = kwargs.pop('atomic', False)

        super().__init__(*args, **kwargs)

        # we have to manually call '__init__' on CoreObject because 'Session'
        # does not call 'super().__init__()' in its '__init__' method.
        CoreObject.__init__(self)

    def __str__(self):
        """
        gets the string representation of current session.

        :rtype: str
        """

        return '{fullname}: atomic={atomic}'.format(fullname=self.get_fully_qualified_name(),
                                                    atomic=self.atomic)

    def execute(self, statement, params=None,
                execution_options=EMPTY_DICT, bind_arguments=None,
                _parent_execute_state=None, _add_event=None, **kw):
        """
        executes a sql expression construct or string statement.

        it will be executed within the current transaction.
        the current transaction will be resolved using the given
        entity class or object or bind name.

        this method is a wrapper for `sqlalchemy.orm.session.execute()` method
        which wraps it to bind to correct engine on different situations.
        for full documentation see the `sqlalchemy.orm.session.execute()` docs.

        :param Executable | str statement: an executable statement `Executable`
                                           expression such as `expression.select`
                                           or string sql statement to be executed.

        :param dict | list[dict] params: optional dictionary, or list of
                                         dictionaries, containing bound
                                         parameter values. if a single dictionary,
                                         single-row execution occurs. if a list of
                                         dictionaries, an `executemany' will be invoked.
                                         the keys in each dictionary must correspond
                                         to parameter names present in the statement.

        :param dict execution_options: optional dictionary of execution options,
                                       which will be associated with the statement execution.
                                       this dictionary can provide a subset of the options that
                                       are accepted by `_future.Connection.execution_options`
                                       method, and may also provide additional options understood
                                       only in an ORM context.

        :param dict bind_arguments: dictionary of additional arguments to determine
                                    the bind. may include `mapper`, `bind` or other custom
                                    arguments. contents of this dictionary are passed to the
                                    `.Session.get_bind` method.

        :param _parent_execute_state: for internal usage.
        :param _add_event: for internal usage.

        :keyword str bind_name: database bind name from `database.binds.ini`
                                file on which given statement must be executed.
                                this keyword argument has precedence over `mapper`
                                and `statement` when locating a bind.

        :keyword bool transient: specifies that raw sql expressions must not affect
                                 the database. if set to True and a raw sql is
                                 given, it could not contain `NON_TRANSIENT_KEYWORDS`.
                                 defaults to False if not provided.

        :raises InvalidDatabaseBindError: invalid database bind error.
        :raises TransientSQLExpressionRequiredError: transient sql expression required error.

        :returns: results of the statement execution.
        :rtype: ResultProxy
        """

        bind = None
        if bind_arguments is not None:
            bind = bind_arguments.get('bind')

        bind_name = kw.pop('bind_name', None)
        if bind is None and bind_name is not None:
            bind = database_services.get_bounded_engines().get(bind_name, None)
            if bind is None:
                raise InvalidDatabaseBindError('Database bind name [{bind_name}] '
                                               'is not available in database.binds '
                                               'config store.'
                                               .format(bind_name=bind_name))

            if bind_arguments is None:
                bind_arguments = dict()

            bind_arguments.update(bind=bind)

        transient = kw.pop('transient', False)
        if transient is True and isinstance(statement, (str, TextClause)):
            raw_sql = None
            if isinstance(statement, TextClause):
                raw_sql = statement.text.lower()
            else:
                raw_sql = statement.lower()

            for item in self.NON_TRANSIENT_KEYWORDS:
                if item in raw_sql:
                    raise TransientSQLExpressionRequiredError(_('Transient sql expressions '
                                                                'should not contain [{keyword}] '
                                                                'keyword.')
                                                              .format(keyword=item.strip()))

        return super().execute(statement, params, execution_options,
                               bind_arguments, _parent_execute_state,
                               _add_event, **kw)

    def get_bind(self, mapper=None, clause=None, bind=None,
                 _sa_skip_events=None, _sa_skip_for_implicit_returning=False):
        """
        returns a `bind` to which this `CoreSession` is bounded.

        this method extends the `get_bind() method of `Session` by
        trying to find bounded engine to tables of a raw string query.

        the `bind` is usually an instance of `Engine`,
        except in the case where the `CoreSession` has been
        explicitly bound directly to a `Connection`.

        for a multiply-bound or unbound `CoreSession`, the
        `mapper` or `clause` arguments are used to determine the
        appropriate bind to return.

        note that the 'mapper' argument is usually present
        when `CoreSession.get_bind` is called via an orm
        operation such as a `Session.query`, each
        individual INSERT/UPDATE/DELETE operation within a
        `Session.flush`, call, etc.

        The order of resolution is:

        1. if mapper given and session.binds is present,
           locate a bind based first on the mapper in use, then
           on the mapped class in use, then on any base classes that are
           present in the `__mro__` of the mapped class, from more specific
           superclasses to more general.
        2. if clause given and session.binds is present,
           locate a bind based on `Table` objects
           found in the given clause present in session.binds.
        3. if session.bind is present, return that.
        4. if clause given, attempt to return a bind
           linked to the `MetaData` ultimately
           associated with the clause.
        5. if mapper given, attempt to return a bind
           linked to the `MetaData` ultimately
           associated with the `Table` or other
           selectable to which the mapper is mapped.
        6. no bind can be found, `sqlalchemy.exc.UnboundExecutionError`
           is raised.

        :param BaseEntity mapper: a BaseEntity instance or class. the bind can be derived from
                                  a `Mapper` first by consulting the `binds` map associated
                                  with this `Session`, and secondly by consulting the
                                  `Metadata` associated with the `Table` to which the
                                  `Mapper` is mapped for a bind.

        :param ClauseElement | str clause: a raw string query or a `ClauseElement`
                                           (i.e. `~.sql.expression.select`,
                                           `~.sql.expression.text`, etc.). if the
                                           `Mapper` argument is not present or could not
                                           produce a bind, the given expression construct
                                           will be searched for a bound element, typically
                                           a `Table` associated with `Metadata`.

        :param Engine bind: if provided, it will be returned immediately.
        :param _sa_skip_events: for internal usage.
        :param _sa_skip_for_implicit_returning: for internal usage.

        :rtype: Engine
        """

        if bind is not None:
            return bind

        if mapper is None and isinstance(clause, (str, TextClause)):
            tables = extractor_services.find_table_names(clause)
            if len(tables) > 0:
                return database_services.get_table_engine(tables[0])

        return super().get_bind(mapper, clause, bind, _sa_skip_events,
                                _sa_skip_for_implicit_returning)

    @property
    def atomic(self):
        """
        gets a value indicating that this session is atomic.

        :rtype: bool
        """

        return self._atomic



# -*- coding: utf-8 -*-
"""
database bulk package.
"""

from pyrin.packaging.base import Package


class DatabaseBulkPackage(Package):
    """
    database bulk package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'database.bulk.component'



# -*- coding: utf-8 -*-
"""
orm types base module.
"""

from abc import abstractmethod

from sqlalchemy import TypeDecorator

from pyrin.core.exceptions import CoreNotImplementedError


class CoreCustomType(TypeDecorator):
    """
    core custom type class.

    all application custom types must be subclassed from this type.
    """

    # the underlying type implementation of this custom type.
    # it must be a subclass of sqlalchemy types.
    impl = None

    @abstractmethod
    def _to_database(self, value, dialect):
        """
        converts given value to be emitted to database.

        this method must be overridden in subclasses.

        :param object value: value to be processed.
        :param Dialect dialect: the dialect in use.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _from_database(self, value, dialect):
        """
        converts given value to python type after fetching it from database.

        :param object value: value to be processed.
        :param Dialect dialect: the dialect in use.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _coerce_to_string(self, value, dialect):
        """
        coerces the given value to string before sending to database.

        subclasses must override this method if they want to use literal params.

        :param object value: value to be processed.
        :param Dialect dialect: the dialect in use.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    def load_dialect_impl(self, dialect):
        """
        returns a `TypeEngine` object corresponding to a dialect.

        :param Dialect dialect: the dialect in use.

        :rtype: TypeEngine
        """

        return super().load_dialect_impl(dialect)

    def compare_against_backend(self, dialect, conn_type):
        """
        returns True if this type is the same as the given database type.

        or None to allow the default implementation to compare these
        types. a return value of False means the given type does not
        match this type.

        :param Dialect dialect: the dialect in use.
        :param TypeEngine conn_type: type of the returned value from database.

        :rtype: bool
        """

        return super().compare_against_backend(dialect, conn_type)

    def process_literal_param(self, value, dialect):
        """
        receives a literal parameter value to be rendered inline within a statement.

        this method is used when the compiler renders a literal value
        without using binds, typically within DDL such as in the `server default`
        of a column or an expression within a CHECK constraint.

        the returned string will be rendered into the output string.

        :param TypeEngine value: data to operate upon. it could be `None`.
        :param Dialect dialect: the dialect in use.

        :rtype: str
        """

        if value is None:
            return value

        result = self._to_database(value, dialect)
        if result is None or isinstance(result, str):
            return result

        return self._coerce_to_string(result, dialect)

    def process_bind_param(self, value, dialect):
        """
        receives a bound parameter value to be converted.

        :param TypeEngine value: data to operate upon. it could be `None`.
        :param Dialect dialect: the dialect in use.

        :rtype: object
        """

        if value is None:
            return value

        return self._to_database(value, dialect)

    def process_result_value(self, value, dialect):
        """
        receives a result-row column value to be converted.

        :param str value: data to operate upon. it could be `None`.
        :param Dialect dialect: the dialect in use.

        :rtype: object
        """

        if value is None:
            return value

        return self._from_database(value, dialect)

    @property
    @abstractmethod
    def python_type(self):
        """
        gets the python type object expected to be returned by instances of this type.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: type
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
api package.
"""

import pyrin.application.services as application_services

from pyrin.packaging.base import Package


class APIPackage(Package):
    """
    api package class.
    """

    NAME = __name__
    DEPENDS = ['pyrin.configuration',
               'pyrin.logging']

    COMPONENT_NAME = 'api.component'
    CONFIG_STORE_NAMES = ['api']

    def _load_configs(self, config_services):
        """
        loads all required configs of this package.
        this method is intended for overriding by
        subclasses to do custom configurations.

        :param Module config_services: configuration services dependency.
                                       to be able to overcome circular dependency problem,
                                       we should inject configuration services dependency
                                       into this method. because all other packages are
                                       referenced `packaging.base` module in them, so we
                                       can't import `pyrin.configuration.services` in this
                                       module. this is more beautiful in comparison to
                                       importing it inside this method.
        """

        flat_configs = config_services.get_all('api')
        application_services.configure(flat_configs)



# -*- coding: utf-8 -*-
"""
security manager module.
"""

from abc import abstractmethod

import pyrin.security.encryption.services as encryption_services
import pyrin.security.hashing.services as hashing_services

from pyrin.core.globals import _
from pyrin.core.structs import Manager
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.security import SecurityPackage
from pyrin.security.exceptions import InvalidPasswordLengthError, \
    InvalidEncryptionTextLengthError


class SecurityManager(Manager):
    """
    security manager class.
    this class is intended to provide some services needed in pyrin application.
    the top level application must extend this class considering business requirements.
    """

    package_class = SecurityPackage

    def get_password_hash(self, password, **options):
        """
        gets the given password's hash.

        :param str password: password to get it's hash.

        :keyword bool is_encrypted: specifies that given password is encrypted.
                                    defaults to False if not provided.

        :raises InvalidPasswordLengthError: invalid password length error.

        :rtype: str
        """

        if password is None or len(password) == 0:
            raise InvalidPasswordLengthError(_('Input password has invalid length.'))

        decrypted_password = password
        is_encrypted = options.get('is_encrypted', False)
        if is_encrypted is True:
            decrypted_password = encryption_services.decrypt(password)

        hashed_password = hashing_services.generate_hash(decrypted_password)
        return hashed_password

    def encrypt(self, text, **options):
        """
        encrypts the given text and returns the encrypted value.

        :param str text: text to be encrypted.

        :raises InvalidEncryptionTextLengthError: invalid encryption text length error.

        :rtype: str
        """

        if text is None or len(text) == 0:
            raise InvalidEncryptionTextLengthError(_('Input text has invalid length.'))

        return encryption_services.encrypt(text)

    @abstractmethod
    def has_permission(self, user, permissions, **options):
        """
        gets a value indicating that given user has the specified permissions.

        :param user: user identity to check its permissions.
        :param list[PermissionBase] permissions: permissions to check for user.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
deserializer handlers boolean module.
"""

import re

from pyrin.converters.deserializer.handlers.base import StringPatternDeserializerBase
from pyrin.converters.deserializer.decorators import deserializer


@deserializer()
class BooleanDeserializer(StringPatternDeserializerBase):
    """
    boolean deserializer class.
    """

    # matches the bool inside string.
    # example: true, false
    # matching is case-insensitive.
    TRUE_REGEX = re.compile(r'^true$', re.IGNORECASE)
    FALSE_REGEX = re.compile(r'^false$', re.IGNORECASE)

    def __init__(self, **options):
        """
        creates an instance of BooleanDeserializer.

        :keyword list[tuple[Pattern, int, int]] accepted_formats: a list of custom accepted
                                                                  formats and their min and
                                                                  max length for boolean
                                                                  deserialization.

        :note accepted_formats: list[tuple[Pattern format, int min_length, int max_length]]

        :keyword bool internal: specifies that this deserializer is internal.
                                internal deserializers will not be used for
                                deserializing client inputs.
                                defaults to False if not provided.
        """

        super().__init__(**options)

        self._converter_map = self._get_converter_map()

    def _deserialize(self, value, **options):
        """
        deserializes the given value.

        :param str value: value to be deserialized.

        :keyword Pattern matching_pattern: the pattern that has matched the value.

        :rtype: bool
        """

        pattern = options.pop('matching_pattern')
        return self._converter_map[pattern]

    @property
    def default_formats(self):
        """
        gets default accepted formats that this deserializer could deserialize value from.

        :returns: list[tuple[Pattern format, int min_length, int max_length]]
        :rtype: list[tuple[Pattern, int, int]]
        """

        return [(self.TRUE_REGEX, 4, 4),
                (self.FALSE_REGEX, 5, 5)]

    def _get_converter_map(self):
        """
        gets converter map dictionary.

        :returns: dict[Pattern format: bool value]
        :rtype: dict
        """

        return {self.TRUE_REGEX: True,
                self.FALSE_REGEX: False}



# -*- coding: utf-8 -*-
"""
babel manager module.
"""

import os

import pyrin.utils.path as path_utils
import pyrin.application.services as application_services

from pyrin.cli.mixin.handler import CLIMixin
from pyrin.core.structs import Manager
from pyrin.globalization.locale.babel import BabelPackage
from pyrin.globalization.locale.babel.enumerations import BabelCLIHandlersEnum
from pyrin.globalization.locale.babel.exceptions import LocaleAlreadyExistedError
from pyrin.globalization.locale.babel.interface import BabelCLIHandlerBase
from pyrin.utils.custom_print import print_warning
from pyrin.utils.exceptions import PathAlreadyExistedError


class BabelManager(Manager, CLIMixin):
    """
    babel manager class.
    """

    _cli_handler_type = BabelCLIHandlerBase
    package_class = BabelPackage

    def enable(self, include_pyrin=True, include_app=True):
        """
        enables locale management for the application.

        :param bool include_pyrin: specifies that it should extract pyrin localizable
                                   messages. defaults to True if not provided.

        :param bool include_app: specifies that it should extract application
                                 localizable messages. defaults to True if not provided.
        """

        locale_path = application_services.get_locale_path()

        try:
            path_utils.create_directory(locale_path)
        except PathAlreadyExistedError:
            print_warning('Locale has been already enabled.', force=True)

        if include_pyrin or include_app:
            self._extract(include_pyrin, include_app)

    def rebuild(self, include_pyrin=True, include_app=True, locale=None):
        """
        it will do the three complete steps needed to
        update and compile locales with new messages.

        it will do:
            1. extract
            2. update
            3. compile

        this command is defined for convenient of usage, but if you need
        to do these steps separately, you could ignore this command and
        use the relevant command for each step.

        :keyword bool include_pyrin: specifies that it must extract pyrin localizable
                                     messages as well. defaults to True if not provided.

        :keyword bool include_app: specifies that it must extract application localizable
                                   messages as well. defaults to True if not provided.

        :keyword str locale: locale name of the catalog to compile.
                             it will compile all catalogs if not provided.
        """

        self._extract(include_pyrin, include_app)
        self.execute(BabelCLIHandlersEnum.UPDATE, locale=locale)
        self.execute(BabelCLIHandlersEnum.COMPILE, locale=locale)

    def _extract(self, include_pyrin=True, include_app=True):
        """
        extracts messages from source files and generates a `.pot` file.

        :keyword bool include_pyrin: specifies that it must extract pyrin localizable
                                     messages as well. defaults to True if not provided.

        :keyword bool include_app: specifies that it must extract application localizable
                                   messages as well. defaults to True if not provided.
        """

        self.execute(BabelCLIHandlersEnum.EXTRACT,
                     include_pyrin=include_pyrin,
                     include_app=include_app)

    def check_init(self, locale):
        """
        checks that locale with given name does not exist.

        :param str locale: locale name for the new localized catalog.
                           for example: `en` or `fr` or ...

        :raises LocaleAlreadyExistedError: locale already existed error.
        """

        if locale not in (None, '') and not locale.isspace():
            locale_path = application_services.get_locale_path()
            full_path = os.path.abspath(os.path.join(locale_path, locale))

            if os.path.exists(full_path):
                raise LocaleAlreadyExistedError('The [{name}] locale is already existed '
                                                'in [{root}]. if you want to regenerate '
                                                'this locale, you should delete the '
                                                '[{dir}] directory first.'
                                                .format(name=locale, root=locale_path,
                                                        dir=full_path))



# -*- coding: utf-8 -*-
"""
token interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.structs import CoreObject, MultiSingletonMeta
from pyrin.core.exceptions import CoreNotImplementedError


class TokenSingletonMeta(MultiSingletonMeta):
    """
    token singleton meta class.
    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractTokenBase(CoreObject, metaclass=TokenSingletonMeta):
    """
    abstract token base class.
    """

    @abstractmethod
    def generate_access_token(self, payload, **options):
        """
        generates an access token from the given inputs and returns it.

        the generated token is in the form of `header_hash.payload_hash.signature_hash`
        and each part is encoded using a signing key.

        :param dict payload: a dictionary containing key/values as payload.
                             note that for better performance, keep payload
                             as small as possible.

        :keyword dict custom_headers: a dictionary containing custom headers.

        :keyword bool is_fresh: indicates that this token is fresh.
                                being fresh means that token is generated by
                                providing user credentials to server.
                                if not provided, defaults to False.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: token.
        :rtype: str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def generate_refresh_token(self, payload, **options):
        """
        generates a refresh token from the given inputs and returns it.

        the generated token is in the form of `header_hash.payload_hash.signature_hash`
        and each part is encoded using a signing key.

        :param dict payload: a dictionary containing key/values as payload.
                             note that for better performance, keep payload
                             as small as possible.

        :keyword dict custom_headers: a dictionary containing custom headers.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: token.
        :rtype: str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_payload(self, token, **options):
        """
        decodes token and gets the payload data.

        if token signature could not be verified, it will raise an error.

        :param str token: token to get it's payload.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: dict
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_unverified_payload(self, token, **options):
        """
        decodes token and gets the payload data without verifying the signature.

        note that returned payload must not be trusted for any critical operations.

        :param str token: token to get it's payload.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: dict
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_unverified_header(self, token, **options):
        """
        gets the header dict of token without verifying the signature.

        note that returned header must not be trusted for critical operations.

        :param str token: token to get it's header.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: dict
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def generate_key(self, **options):
        """
        generates a valid key for this handler and returns it.

        :keyword int length: the length of generated key in bytes.
                             note that some token handlers may not accept custom
                             key length so this value would be ignored on those handlers.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str | tuple[str, str]
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_kid(self):
        """
        gets kid value to be used in token header for this handler.

        it must be unique for each handler.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
audit api module.
"""

import pyrin.audit.services as audit_services

from pyrin.api.router.decorators import api
from pyrin.core.enumerations import HTTPMethodEnum


audit_config = audit_services.get_audit_configurations()
is_enabled = audit_config.pop('enabled', False)

if is_enabled is True:
    @api(**audit_config, methods=HTTPMethodEnum.GET, no_cache=True)
    def inspect(**options):
        """
        inspects all registered packages and gets inspection data.

        it includes all available info, but it could be customized using different options.
        other custom implementations could also be excluded from result if the related name
        of their hook is passed with False value. for example `database=False`.

        it returns a tuple of two values. the first value is a dict containing the
        inspection data, and the second value is the related status code for the response.
        in case of any inspection has been failed, the status code will be internal
        server error 500.

        :keyword bool application: specifies that application info must be included.
        :keyword bool packages: specifies that loaded packages info must be included.
        :keyword bool framework: specifies that framework info must be included.
        :keyword bool python: specifies that python info must be included.
        :keyword bool os: specifies that operating system info must be included.
        :keyword bool hardware: specifies that hardware info must be included.

        :keyword bool traceback: specifies that on failure report, it must include
                                 the traceback of errors.
                                 defaults to True if not provided.

        :returns: tuple[dict(dict application: application info,
                             dict packages: loaded packages info,
                             dict framework: framework info,
                             dict python: python info,
                             dict platform: platform info),
                        int status_code]
        :rtype: tuple[dict, int]
        """

        return audit_services.inspect(**options)



# -*- coding: utf-8 -*-
"""
configuration component module.
"""

from pyrin.application.decorators import component
from pyrin.configuration import ConfigurationPackage
from pyrin.configuration.manager import ConfigurationManager
from pyrin.application.structs import Component


@component(ConfigurationPackage.COMPONENT_NAME)
class ConfigurationComponent(Component, ConfigurationManager):
    """
    configuration component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
utils number module.
"""

from decimal import Decimal


def coerce_to_int(value):
    """
    gets the integer equivalent of given decimal or float value.

    if the real value does not have an int equivalent, it returns the same input.

    for example:

    coerce_to_int(1.2) -> 1.2
    coerce_to_int(1.0) -> 1
    coerce_to_int(1) -> 1
    coerce_to_int('1') -> '1'

    :param float | Decimal value: value to be coerced.

    :rtype: int | object
    """

    if not isinstance(value, (float, Decimal)):
        return value

    integer = int(value)
    if integer == value:
        return integer

    return value


def coerce_to_float(value):
    """
    gets the float equivalent of given decimal or integer value.

    if the real value does not have a float equivalent, it returns the same input.

    for example:

    coerce_to_float(1.2) -> 1.2
    coerce_to_float(1.0) -> 1.0
    coerce_to_float(1) -> 1.0
    coerce_to_float('1.2') -> '1.2'

    :param int | Decimal value: value to be coerced.

    :rtype: float | object
    """

    if not isinstance(value, (int, Decimal)) or isinstance(value, bool):
        return value

    return float(value)


def coerce_to_decimal(value):
    """
    gets the decimal equivalent of given float or integer value.

    if the real value does not have a decimal equivalent, it returns the same input.

    for example:

    coerce_to_decimal(1.2) -> 1.1999999
    coerce_to_decimal(1.0) -> 1.0
    coerce_to_decimal(1) -> 1.0
    coerce_to_decimal('1') -> '1'

    :param int | float value: value to be coerced.

    :rtype: Decimal | object
    """

    if not isinstance(value, (int, float)) or isinstance(value, bool):
        return value

    return Decimal(value)



# -*- coding: utf-8 -*-
"""
caching mixin base module.
"""

from abc import abstractmethod

import pyrin.utils.function as func_utils
import pyrin.security.session.services as session_services

from pyrin.caching.structs import CacheableDict
from pyrin.core.exceptions import CoreNotImplementedError


class CacheMixinBase:
    """
    cache mixin base class.
    """

    # each subclass must set this to a dict to get its own container.
    _container = None

    @classmethod
    def set_cache(cls, value, *keys, **options):
        """
        sets given value into the cache.

        :param object value: value to be cached.
        :param object keys: arguments to be used as cache key.
                            all arguments must be hashable.
        """

        key = cls.generate_key(*keys, **options)
        cls._container[key] = value

    @classmethod
    def get_cache(cls, *keys, **options):
        """
        gets the value of given key from cache.

        :param object keys: arguments to be used as cache key.
                            all arguments must be hashable.

        :returns: object
        """

        key = cls.generate_key(*keys, **options)
        return cls._container.get(key)

    @classmethod
    def clear_cache(cls):
        """
        clears all cached values.
        """

        cls._container.clear()

    @classmethod
    @abstractmethod
    def generate_key(cls, *keys, **options):
        """
        generates a cache key from given inputs.

        :param object keys: arguments to be used as cache key.
                            all arguments must be hashable.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: object
        """

        raise CoreNotImplementedError()


class SimpleKeyGeneratorMixin:
    """
    simple key generator mixin class.
    """

    def generate_key(self, func, parent, *args, **options):
        """
        generates a cache key from given inputs.

        :param function func: function to to be cached.
        :param type | object parent: parent class or instance of given function.

        :returns: hash of generated key
        :rtype: int
        """

        return hash(self._generate_key(func, parent, *args, **options))

    def _generate_key(self, func, parent, *args, **options):
        """
        generates a cache key from given inputs.

        :param function func: function to to be cached.
        :param type | object parent: parent class or instance of given function.

        :returns: tuple[type parent, str function_name]
        :rtype: tuple[type, str]
        """

        parent_type = None
        name = func.__name__
        if parent is None:
            name = func_utils.get_fully_qualified_name(func)
        else:
            parent_type = self._get_parent_type(parent)

        return parent_type, name

    def _get_parent_type(self, parent):
        """
        gets the parent type from given input.

        if it is a class itself, it returns the same input.

        :param type | object parent: class or instance to get its type.

        :rtype: type
        """

        parent_type = parent
        if not isinstance(parent_type, type):
            parent_type = type(parent_type)

        return parent_type


class ComplexKeyGeneratorMixin(SimpleKeyGeneratorMixin):
    """
    complex key generator mixin class.
    """

    def generate_key(self, func, inputs, kw_inputs, *args, **options):
        """
        generates a cache key from given inputs.

        :param function func: function to to be cached.
        :param tuple inputs: function positional arguments.
        :param dict kw_inputs: function keyword arguments.

        :keyword bool consider_user: specifies that current user must be included in
                                     key generation. it will be get from `caching` config
                                     store if not provided.

        :returns: hash of generated key
        :rtype: int
        """

        return hash(self._generate_key(func, inputs, kw_inputs, *args, **options))

    def _generate_key(self, func, inputs, kw_inputs, *args, **options):
        """
        generates a cache key from given inputs.

        :param function func: function to to be cached.
        :param tuple inputs: function positional arguments.
        :param dict kw_inputs: function keyword arguments.

        :keyword bool consider_user: specifies that current user must be included in
                                     key generation. it will be get from `caching` config
                                     store if not provided.

        :returns: tuple[type parent, str function_name,
                        dict inputs, object user,
                        object component key,
                        str timezone, str locale]

        :rtype: tuple[type, str, dict, object, object, str, str]
        """

        current_request = session_services.get_safe_current_request()
        timezone = None
        locale = None
        current_user = None
        component_key = None

        if current_request is not None:
            consider_user = options.get('consider_user', self.consider_user)
            if consider_user is not False:
                current_user = current_request.cacheable_user

            component_key = current_request.component_custom_key
            timezone = current_request.timezone.zone
            locale = current_request.locale

        cacheable_inputs, parent = func_utils.get_inputs(func, inputs,
                                                         kw_inputs, CacheableDict)

        parent_type, name = super()._generate_key(func, parent, *args, **options)

        return parent_type, name, cacheable_inputs, \
            current_user, component_key, timezone, locale



# -*- coding: utf-8 -*-
"""
alembic cli module.
"""

import pyrin.database.migration.alembic.services as alembic_services

from pyrin.cli.decorators import cli, cli_invoke, cli_group
from pyrin.core.structs import CLI


@cli_group('alembic')
class AlembicCLI(CLI):
    """
    alembic cli class.

    this class exposes all alembic cli commands.
    """

    _execute_service = alembic_services.execute

    @cli_invoke
    def enable(self, **options):
        """
        enable migrations for application.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """

        return alembic_services.enable()

    @cli
    def branches(self, **options):
        """
        show current branch points.

        :keyword bool verbose: use more verbose output.
                               defaults to False if not provided.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """
        pass

    @cli
    def current(self, **options):
        """
        display the current revision for a database.

        :keyword bool verbose: use more verbose output.
                               defaults to False if not provided.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """
        pass

    @cli
    def downgrade(self, revision, **options):
        """
        revert to a previous version.
        use `base` to revert all revisions.

        :param str revision: revision identifier.

        :keyword bool sql: don't emit sql to database, dump to standard
                           output/file instead. see docs on offline mode.
                           defaults to False if not provided.

        :keyword str tag: arbitrary `tag` name. can be used
                          by custom `env.py` scripts.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """
        pass

    @cli
    def heads(self, **options):
        """
        show current available heads in the script directory.

        :keyword bool resolve_dependencies: treat dependency versions as down revisions.
                                            defaults to False if not provided.

        :keyword bool verbose: use more verbose output.
                               defaults to False if not provided.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """
        pass

    @cli
    def history(self, **options):
        """
        list change-set scripts in chronological order.

        :keyword str revision_range: specify a revision range.
                                     format is [start]:[end]

        :keyword bool indicate_current: indicate the current revision.
                                        defaults to False if not provided.

        :keyword bool verbose: use more verbose output.
                               defaults to False if not provided.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """
        pass

    @cli
    def merge(self, revisions, **options):
        """
        merge two revisions together. creates a new migration file.

        :param str | list[str] revisions: one or more revisions,
                                          or `heads` for all heads.

        :keyword str message: message string to use with `revision`.
                              defaults to current timestamp if not provided.

        :keyword str branch_label: specify a branch label to
                                   apply to the new revision.

        :keyword str revision_id: specify a hardcoded revision
                                  id instead of generating one.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """
        pass

    @cli
    def revision(self, **options):
        """
        create a new revision file.

        :keyword str message: message string to use with `revision`.
                              defaults to current timestamp if not provided.

        :keyword bool autogenerate: populate revision script with candidate migration
                                    operations, based on comparison of database to model.
                                    defaults to True if not provided.

        :keyword bool sql: don't emit sql to database, dump to standard
                           output/file instead. see docs on offline mode.
                           defaults to False if not provided.

        :keyword str head: specify head revision or <branchname>@head
                           to base new revision on.

        :keyword bool splice: allow a non-head revision as the `head` to splice onto.
                              defaults to False if not provided.

        :keyword str branch_label: specify a branch label to
                                   apply to the new revision.

        :keyword str version_path: specify specific path from config for version file.
        :keyword str revision_id: specify a hardcoded revision id instead of generating one.

        :keyword str | list[str] depends_on: specify one or more revision identifiers
                                             which this revision should depend on.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """
        pass

    @cli
    def show(self, revision, **options):
        """
        show the revision(s) denoted by the given symbol.

        :param str revision: revision identifier.
                             it could be a part of revision identifier
                             to filter all matching revisions.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """
        pass

    @cli
    def stamp(self, revisions, **options):
        """
        stamp the revision table with the given revision(s).
        don't run any migrations.

        :param str | list[str] revisions: one or more revisions,
                                          or `heads` for all heads.

        :keyword bool sql: don't emit sql to database, dump to standard
                           output/file instead. see docs on offline mode.
                           defaults to False if not provided.

        :keyword str tag: arbitrary `tag` name. can be used
                          by custom `env.py` scripts.

        :keyword bool purge: unconditionally erase the version table before stamping.
                             defaults to False if not provided.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """
        pass

    @cli
    def upgrade(self, revision, **options):
        """
        upgrade to a later version.
        use `head` to upgrade to latest revision.

        :param str revision: revision identifier.

        :keyword bool sql: don't emit sql to database, dump to standard
                           output/file instead. see docs on offline mode.
                           defaults to False if not provided.

        :keyword str tag: arbitrary `tag` name. can be used
                          by custom `env.py` scripts.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """
        pass



# -*- coding: utf-8 -*-
"""
response status services module.
"""

from pyrin.processor.response.status import ResponseStatusPackage
from pyrin.application.services import get_component


def get_status_code(method=None, **options):
    """
    gets the corresponding status code for given http method.

    if method is not provided, it gets the status code for current request's method.

    :param str method: http method to get its corresponding status code.
    :enum method:
        GET = 'GET'
        POST = 'POST'
        PUT = 'PUT'
        DELETE = 'DELETE'
        HEAD = 'HEAD'
        TRACE = 'TRACE'
        OPTIONS = 'OPTIONS'
        PATCH = 'PATCH'
        COPY = 'COPY'
        LINK = 'LINK'
        UNLINK = 'UNLINK'
        PURGE = 'PURGE'
        VIEW = 'VIEW'

    :keyword bool has_content: specifies that the response body has content.
                               if set to False, the status code returned will
                               be `NO_CONTENT=204` for all http methods.
                               defaults to True if not provided.

    :rtype: int
    """

    return get_component(ResponseStatusPackage.COMPONENT_NAME).get_status_code(method,
                                                                               **options)


def is_client_error(status_code, **options):
    """
    gets a value indicating that given status code is a client error.

    if returns True if the provided status code is
    from `ClientErrorResponseCodeEnum` values.

    :param int status_code: status code to be checked.

    :keyword bool strict_status: specifies that it should only consider
                                 the status code as client error if it is from
                                 `ClientErrorResponseCodeEnum` values. otherwise
                                 all codes from `CLIENT_ERROR_CODE_MIN` to
                                 `CLIENT_ERROR_CODE_MAX` will be considered as
                                 client error. defaults to True if not provided.

    :rtype: bool
    """

    return get_component(ResponseStatusPackage.COMPONENT_NAME).is_client_error(status_code,
                                                                               **options)


def is_server_error(status_code, **options):
    """
    gets a value indicating that given status code is a server error.

    if returns True if the provided status code is
    from `ServerErrorResponseCodeEnum` values.

    :param int status_code: status code to be checked.

    :keyword bool strict_status: specifies that it should only consider
                                 the status code as server error if it is from
                                 `ServerErrorResponseCodeEnum` values. otherwise
                                 all codes from `SERVER_ERROR_CODE_MIN` to
                                 `SERVER_ERROR_CODE_MAX` will be considered
                                 as server error. defaults to True if not provided.

    :rtype: bool
    """

    return get_component(ResponseStatusPackage.COMPONENT_NAME).is_server_error(status_code,
                                                                               **options)


def is_error(status_code, **options):
    """
    gets a value indicating that given status code is a client or server error.

    if returns True if the provided status code is from
    `ClientErrorResponseCodeEnum` or `ServerErrorResponseCodeEnum` values.

    :param int status_code: status code to be checked.

    :keyword bool strict_status: specifies that it should only consider
                                 the status code as error if it is from
                                 `ClientErrorResponseCodeEnum` or
                                 `ServerErrorResponseCodeEnum` values. otherwise
                                 all codes from `CLIENT_ERROR_CODE_MIN` to
                                 `CLIENT_ERROR_CODE_MAX` or from
                                 `SERVER_ERROR_CODE_MIN` to `SERVER_ERROR_CODE_MAX`
                                 will be considered as error.
                                 defaults to True if not provided.

    :rtype: bool
    """

    return get_component(ResponseStatusPackage.COMPONENT_NAME).is_error(status_code,
                                                                        **options)


def is_information(status_code, **options):
    """
    gets a value indicating that given status code is a information code.

    if returns True if the provided status code is
    from `InformationResponseCodeEnum` values.

    :param int status_code: status code to be checked.

    :keyword bool strict_status: specifies that it should only consider
                                 the status code as information if it is from
                                 `InformationResponseCodeEnum` values. otherwise
                                 all codes from `INFORMATION_CODE_MIN` to
                                 `INFORMATION_CODE_MAX` will be considered as
                                 information. defaults to True if not provided.

    :rtype: bool
    """

    return get_component(ResponseStatusPackage.COMPONENT_NAME).is_information(status_code,
                                                                              **options)


def is_success(status_code, **options):
    """
    gets a value indicating that given status code is a success code.

    if returns True if the provided status code is
    from `SuccessfulResponseCodeEnum` values.

    :param int status_code: status code to be checked.

    :keyword bool strict_status: specifies that it should only consider
                                 the status code as success if it is from
                                 `SuccessfulResponseCodeEnum` values. otherwise
                                 all codes from `SUCCESS_CODE_MIN` to
                                 `SUCCESS_CODE_MAX` will be considered as success.
                                 defaults to True if not provided.

    :rtype: bool
    """

    return get_component(ResponseStatusPackage.COMPONENT_NAME).is_success(status_code,
                                                                          **options)


def is_redirection(status_code, **options):
    """
    gets a value indicating that given status code is a redirection code.

    if returns True if the provided status code is
    from `RedirectionResponseCodeEnum` values.

    :param int status_code: status code to be checked.

    :keyword bool strict_status: specifies that it should only consider
                                 the status code as redirection if it is from
                                 `RedirectionResponseCodeEnum` values. otherwise
                                 all codes from `REDIRECTION_CODE_MIN` to
                                 `REDIRECTION_CODE_MAX` will be considered
                                 as redirection. defaults to True if not provided.

    :rtype: bool
    """

    return get_component(ResponseStatusPackage.COMPONENT_NAME).is_redirection(status_code,
                                                                              **options)


def is_processed(status_code, **options):
    """
    gets a value indicating that given status code is a processed code.

    processed codes are information or success or redirection code.
    if returns True if the provided status code is
    from `InformationResponseCodeEnum` or `SuccessfulResponseCodeEnum`
    or `RedirectionResponseCodeEnum` values.

    :param int status_code: status code to be checked.

    :keyword bool strict_status: specifies that it should only consider
                                 the status code as processed if it is from
                                 `InformationResponseCodeEnum` or
                                 `SuccessfulResponseCodeEnum` or
                                 `RedirectionResponseCodeEnum` values. otherwise
                                 all codes from `INFORMATION_CODE_MIN`
                                 to `INFORMATION_CODE_MAX` or from
                                 `SUCCESS_CODE_MIN` to `SUCCESS_CODE_MAX`
                                 or from `REDIRECTION_CODE_MIN` to
                                 `REDIRECTION_CODE_MAX` will be considered
                                 as processed. defaults to True if not provided.

    :rtype: bool
    """

    return get_component(ResponseStatusPackage.COMPONENT_NAME).is_processed(status_code,
                                                                            **options)



# -*- coding: utf-8 -*-
"""
permission base module.
"""

from abc import abstractmethod

import pyrin.security.permission.services as permission_services

from pyrin.core.structs import CoreObject
from pyrin.core.exceptions import CoreNotImplementedError


class PermissionBase(CoreObject):
    """
    permission base class.

    all application permissions must be subclassed from this.
    """

    def __init__(self, *args, **options):
        """
        initializes an instance of PermissionBase.

        input parameters of this method in subclasses must be
        customized based on application's design requirements.
        """

        super().__init__()
        permission_services.register_permission(self, **options)

    def __hash__(self):
        """
        gets the hash value of current permission.

        :rtype: int
        """

        return hash(self.get_id())

    def __eq__(self, other):
        """
        gets the comparison between current and other permission for equality.

        :param PermissionBase other: other permission instance to be
                                     compared to the current one.

        :rtype: bool
        """

        if not isinstance(other, PermissionBase):
            return False

        return other.get_id() == self.get_id()

    def __ne__(self, other):
        """
        gets the comparison between current and other permission for not equality.

        :param PermissionBase other: other permission instance to be
                                     compared to the current one.

        :rtype: bool
        """

        return not self == other

    @abstractmethod
    def __str__(self):
        """
        gets the string representation of current permission.

        this method must be implemented in subclasses.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    def __repr__(self):
        """
        gets the string representation of current permission.

        :rtype: str
        """

        return str(self)

    @abstractmethod
    def to_entity(self):
        """
        gets the equivalent entity of current permission.

        this method must be implemented in subclasses.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: BaseEntity
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_id(self):
        """
        gets permission id.

        this method must be implemented in subclasses.
        it could return a single value or a combination of multiple values
        (ex. a tuple). note that the returned value must be fully unique for
        each different permission and also it must be a hashable value to
        be used as dict key.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: object
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
database enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class FKRestrictionEnum(CoreEnum):
    """
    fk restriction enum.
    """

    CASCADE = 'CASCADE'
    DELETE = 'DELETE'
    RESTRICT = 'RESTRICT'


class DialectEnum(CoreEnum):
    """
    dialect enum.
    """

    SQLITE = 'sqlite'
    POSTGRESQL = 'postgresql'
    SQLSERVER = 'mssql'
    ORACLE = 'oracle'
    MYSQL = 'mysql'
    SYBASE = 'sybase'
    FIREBIRD = 'firebird'



# -*- coding: utf-8 -*-
"""
caching local items base module.
"""

from copy import deepcopy

import time

from pyrin.core.structs import CoreObject


class LocalCacheItemBase(CoreObject):
    """
    local cache item base class.

    this type of cache item does not support expire time.
    it also keeps the original value into the cache to gain performance.

    all application cache items must be subclassed from this.
    """

    def __init__(self, key, value, *args, **options):
        """
        initializes an instance of LocalCacheItemBase.

        :param object key: hashable object for cache key.
        :param object value: value to be cached.
        """

        super().__init__()

        self._created_on = time.time() * 1000
        self._key = key
        self._value = self._prepare_cache(value)

    def __str__(self):
        """
        gets the string representation of this cache item.

        :rtype: str
        """
        return 'CacheItem: [{key}-{value}]'.format(key=self._key, value=self._value)

    def __repr__(self):
        """
        gets the string representation of this cache item.

        :rtype: str
        """

        return str(self)

    def __hash__(self):
        """
        gets the hash of this cache item.

        :rtype: int
        """

        return hash(self._key)

    def _get_cached_value(self, value):
        """
        gets the cached value.

        it is intended to be overridden in subclasses.

        :param object value: value to be returned from cached.

        :rtype: object
        """

        return value

    def _prepare_cache(self, value):
        """
        prepares value to be cached.

        it is intended to be overridden in subclasses.

        :param object value: value to be cached.

        :rtype: object
        """

        return value

    @property
    def value(self):
        """
        gets the cached value of this item.

        :rtype: object
        """

        return self._get_cached_value(self._value)

    @property
    def key(self):
        """
        gets the key of this item.

        :rtype: object
        """

        return self._key


class ComplexLocalCacheItemBase(LocalCacheItemBase):
    """
    complex local cache item base class.

    this type of cache item supports expire time.
    it also keeps the deep copy of the value into the cache.

    all application complex cache items must be subclassed from this.
    """

    def __init__(self, key, value, expire, **options):
        """
        initializes an instance of ComplexLocalCacheItemBase.

        :param object key: hashable object for cache key.
        :param object value: value to be cached.

        :keyword bool refreshable: specifies that this item's expire time must be
                                   extended on each hit. defaults to False if not
                                   provided.
        """

        super().__init__(key, value, **options)

        self._refreshed_on = self._created_on
        self._expire = expire
        self._refreshable = options.get('refreshable', False)

    def _get_cached_value(self, value):
        """
        gets the cached value.

        :param object value: value to be returned from cached.

        :rtype: object
        """

        if self._refreshable is True and self.is_expired is False:
            self.refresh()

        return deepcopy(value)

    def _prepare_cache(self, value):
        """
        prepares value to be cached.

        :param object value: value to be cached.

        :rtype: object
        """

        return deepcopy(value)

    def refresh(self):
        """
        refreshes the current item to extend its expire time.
        """

        self._refreshed_on = time.time() * 1000

    @property
    def is_expired(self):
        """
        gets a value indicating that this item has been expired.

        :rtype: bool
        """

        return time.time() * 1000 - self._refreshed_on > self._expire

    @property
    def expire(self):
        """
        gets the expire time value of this item in milliseconds.

        :rtype: int
        """

        return self._expire



# -*- coding: utf-8 -*-
"""
orm session exceptions module.
"""

from pyrin.core.exceptions import CoreException


class ORMSessionException(CoreException):
    """
    orm session exception.
    """
    pass


class TransientSQLExpressionRequiredError(ORMSessionException):
    """
    transient sql expression required error.
    """
    pass



# -*- coding: utf-8 -*-
"""
celery audit tasks module.
"""

import pyrin.task_queues.celery.audit.services as celery_audit_services

from pyrin.task_queues.celery.decorators import task


@task
def audit_task():
    """
    this method will be used to check status of celery.
    """

    celery_audit_services.perform_job()



# -*- coding: utf-8 -*-
"""
core exceptions module.
"""

from pyrin.core.enumerations import ServerErrorResponseCodeEnum, ClientErrorResponseCodeEnum


class CoreException(Exception):
    """
    base class for all application exceptions.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of CoreException.

        :keyword dict data: extra data for exception.
        """

        super().__init__(*args)
        self._data = kwargs.get('data', None) or {}
        self._traceback = None
        self._code = ServerErrorResponseCodeEnum.INTERNAL_SERVER_ERROR
        self._description = str(self)

    @property
    def code(self):
        """
        gets the error code.

        :rtype: int
        """

        return self._code

    @property
    def data(self):
        """
        gets the error data.

        :rtype: dict
        """

        return self._data

    @property
    def traceback(self):
        """
        gets the traceback of this exception.

        :rtype: object
        """

        return self._traceback

    @property
    def description(self):
        """
        gets the error description.

        :rtype: str
        """

        return self._description


class CoreBusinessException(CoreException):
    """
    base class for all application business exceptions.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of CoreBusinessException.

        :keyword dict data: extra data for exception.
        """

        super().__init__(*args, **kwargs)
        self._code = ClientErrorResponseCodeEnum.UNPROCESSABLE_ENTITY


class CoreAttributeError(CoreException, AttributeError):
    """
    core attribute error.
    """
    pass


class ContextAttributeError(CoreAttributeError):
    """
    context attribute error.
    """
    pass


class CoreNotImplementedError(CoreException, NotImplementedError):
    """
    core not implemented error.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of CoreNotImplementedError.

        :keyword dict data: extra data for exception.
        """

        super().__init__('This method does not have an implementation.',
                         *args, **kwargs)


class CoreTypeError(CoreException, TypeError):
    """
    core type error.
    """
    pass


class CoreValueError(CoreException, ValueError):
    """
    core value error.
    """
    pass


class CoreKeyError(CoreException, KeyError):
    """
    core key error.
    """
    pass


class CoreAssertionError(CoreException, AssertionError):
    """
    core assertion error.
    """
    pass


class CoreNotADirectoryError(CoreException, NotADirectoryError):
    """
    core not a directory error.
    """
    pass


class CoreFileNotFoundError(CoreException, FileNotFoundError):
    """
    core file not found error.
    """
    pass


class CoreNameError(CoreException, NameError):
    """
    core name error.
    """
    pass


class InvalidHookTypeError(CoreException):
    """
    invalid hook type error.
    """
    pass


class PackageClassIsNotSetError(CoreException):
    """
    package class is not set error.
    """
    pass



# -*- coding: utf-8 -*-
"""
caching local handlers base module.
"""

import pickle

from abc import abstractmethod
from threading import Thread

import pyrin.database.bulk.services as bulk_services
import pyrin.logging.services as logging_services
import pyrin.configuration.services as config_services
import pyrin.globalization.datetime.services as datetime_services

from pyrin.caching.mixin.base import ComplexKeyGeneratorMixin, SimpleKeyGeneratorMixin
from pyrin.caching.models import CacheItemEntity
from pyrin.caching.globals import NO_LIMIT
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.caching.local.items.base import LocalCacheItemBase, ComplexLocalCacheItemBase
from pyrin.database.services import get_current_store
from pyrin.caching.local.containers.base import LocalCacheContainerBase
from pyrin.core.globals import SECURE_FALSE
from pyrin.caching.interface import AbstractCache, AbstractComplexLocalCache, \
    AbstractExtendedLocalCache
from pyrin.caching.exceptions import CacheIsNotPersistentError, CacheNameIsRequiredError, \
    InvalidCacheLimitError, InvalidCacheExpireTimeError
from pyrin.caching.local.handlers.exceptions import InvalidCacheContainerTypeError, \
    InvalidCacheItemTypeError, InvalidCacheClearCountError, InvalidChunkSizeError, \
    CacheClearanceLockTypeIsRequiredError, CacheVersionIsRequiredError, \
    CachePersistentLockTypeIsRequiredError


class LocalCacheBase(SimpleKeyGeneratorMixin, AbstractCache):
    """
    local cache base class.

    this type of caches does not consider method inputs, current user
    and component key in key generation. it only considers the class
    type of function and function name itself. this is useful for caching
    items that never change after application startup and are independent
    from different scoped or global variables.

    it also does not support expire time and size limit for cached values.
    its values are permanent unless manually removed if required.

    it also keeps the real value in the cache, not a deep copy of it to gain
    performance.

    it also does not provide statistic info about hit or missed
    caches, to gain performance.
    """

    # cache name to be used for this cache.
    # it must be unique between all caches.
    cache_name = None

    # a class type to be used as cache container.
    # it could be overridden in subclasses.
    container_class = None

    # a class type to hold each item in the cache.
    # it could be overridden in subclasses.
    cache_item_class = None

    LOGGER = logging_services.get_logger('caching.local')

    def __init__(self, *args, **options):
        """
        initializes an instance of LocalCacheBase.

        :raises CacheNameIsRequiredError: cache name is required error.
        :raises InvalidCacheContainerTypeError: invalid cache container type error.
        :raises InvalidCacheItemTypeError: invalid cache item type error.
        """

        super().__init__()

        if self.cache_name in (None, '') or self.cache_name.isspace():
            raise CacheNameIsRequiredError('Cache name must be provided for '
                                           'cache [{name}].'.format(name=self))

        self._set_name(self.cache_name)

        if self.container_class is None or \
                not issubclass(self.container_class, LocalCacheContainerBase):
            raise InvalidCacheContainerTypeError('Provided cache container [{container}] '
                                                 'for cache [{name}] is not a subclass '
                                                 'of [{base}].'
                                                 .format(container=self.container_class,
                                                         name=self.get_name(),
                                                         base=LocalCacheContainerBase))

        if self.cache_item_class is None or \
                not issubclass(self.cache_item_class, LocalCacheItemBase):
            raise InvalidCacheItemTypeError('Provided cache item [{item}] for cache '
                                            '[{name}] is not a subclass of [{base}].'
                                            .format(item=self.cache_item_class,
                                                    name=self.get_name(),
                                                    base=LocalCacheItemBase))

        self._container = self.container_class()
        self._last_cleared_time = datetime_services.now()

    def _get_cache_item(self, value, *args, **options):
        """
        gets the equivalent cache item for given value.

        :param object value: value to be cached.

        :rtype: CacheItemBase
        """

        return self.cache_item_class(value, *args, **options)

    def _get_configs(self):
        """
        gets the configs of this cache from `caching` config store.

        first, it checks if a section with the name of this cache
        is present in config store. if so, it returns the values from it.
        otherwise it returns the result of `_get_default_configs` method.

        :rtype: dict
        """

        if self.get_name() in config_services.get_section_names('caching'):
            return config_services.get_section('caching', self.get_name())
        else:
            return self._get_default_configs()

    @abstractmethod
    def _get_default_configs(self):
        """
        gets the defaults configs of this cache from `caching` config store.

        this method must be overridden in subclasses to provide custom configs if
        a section with the cache name is not present in config store.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: dict
        """

        raise CoreNotImplementedError('This method must be implemented in '
                                      '[{class_name}] class to provide default '
                                      'configs for cache [{name}].'
                                      .format(class_name=self, name=self.get_name()))

    def _try_set(self, value, func, parent, *args, **options):
        """
        sets a new value into cached items.

        this method will generate cache key from given type and function.

        :param object value: value to be cached.
        :param function func: function to cache its result.
        :param type | object parent: parent class or instance of given function.
        """

        key = self.generate_key(func, parent, *args, **options)
        self.set(key, value, **options)

    def _try_get(self, func, parent, *args, default=None, **options):
        """
        gets the value from cache.

        this method will generate cache key from given type and function.
        if key does not exist, it returns None or the specified default value.

        :param function func: function to to get its result.
        :param type | object parent: parent class or instance of given function.
        :param object default: value to be returned if key is not present.

        :returns: object
        """

        key = self.generate_key(func, parent, *args, **options)
        return self.get(key, default=default, **options)

    def set(self, key, value, *args, **options):
        """
        sets a new value into cached items.

        :param object key: hashable key of the cache to be registered.
        :param object value: value to be cached.
        """

        hash_key = hash(key)
        cache_item = self._get_cache_item(hash_key, value, *args, **options)
        self._container[hash_key] = cache_item

    def get(self, key, default=None, **options):
        """
        gets the value from cache.

        if key does not exist, it returns None or the specified default value.

        :param object key: hashable key to get its value from cache.
        :param object default: value to be returned if key is not present.

        :returns: object
        """

        result = self._container.get(hash(key), default)
        if result not in (None, default):
            return result.value

        return result

    def try_set(self, value, func, parent, *args, **options):
        """
        sets a new value into cached items.

        this method will generate cache key from given type and function.
        if the provided values are not hashable, this method won't raise
        an error and logs it silently.

        :param object value: value to be cached.
        :param function func: function to cache its result.
        :param type | object parent: parent class or instance of given function.
        """

        try:
            self._try_set(value, func, parent, *args, **options)
        except TypeError as error:
            self.LOGGER.exception(str(error))

    def try_get(self, func, parent, *args, default=None, **options):
        """
        gets the value from cache.

        this method will generate cache key from given type and function.
        if key does not exist, it returns None or the specified default value.
        if the provided values are not hashable, this method won't raise
        an error and logs it silently.

        :param function func: function to to get its result.
        :param type | object parent: parent class or instance of given function.
        :param object default: value to be returned if key is not present.

        :returns: object
        """

        try:
            return self._try_get(func, parent, *args, default=default, **options)
        except TypeError as error:
            self.LOGGER.exception(str(error))
            return None

    def contains(self, key):
        """
        gets a value indicating that given key is existed in the cached items.

        :param object key: key to be checked for existence.

        :rtype: bool
        """

        return hash(key) in self._container

    def pop(self, key, default=None):
        """
        pops the given key from cache and returns its value.

        if key does not exist, it returns None or the specified default value.

        :param object key: key to get its value.
        :param object default: value to be returned if key is not present.

        :returns: object
        """

        return self._container.pop(hash(key), default)

    def remove(self, key, **options):
        """
        removes the given key from cache.

        it does nothing if the key is not present in the cache.

        :param object key: key to be removed.
        """

        try:
            del self._container[hash(key)]
        except Exception:
            pass

    def clear(self):
        """
        clears all items from cache.
        """

        self._container.clear()
        self._last_cleared_time = datetime_services.now()

    def items(self):
        """
        gets an iterable of all keys and their values in the cache.

        :returns: iterable[tuple[object, object]]
        """

        return self._container.items()

    def keys(self):
        """
        gets all keys of current cache.

        :returns: iterable[object]
        """

        return self._container.keys()

    def values(self):
        """
        gets all values of current cache.

        :returns: iterable[object]
        """

        return self._container.values()

    @property
    def count(self):
        """
        gets the count of items of this cache.

        :rtype: int
        """

        return len(self._container)

    @property
    def last_cleared_time(self):
        """
        gets the last time in which this cache has been cleared.

        :rtype: datetime.datetime
        """

        return self._last_cleared_time

    @property
    def stats(self):
        """
        get the statistic info about cached items.

        :returns: dict(int count: items count,
                       datetime last_cleared_time: last cleared time,
                       bool persistent: persistent cache)
        :rtype: dict
        """

        return dict(count=self.count,
                    last_cleared_time=self.last_cleared_time,
                    persistent=self.persistent)

    @property
    def persistent(self):
        """
        gets a value indicating that cached items must be persisted to database on shutdown.

        :rtype: bool
        """

        return False


class ExtendedLocalCacheBase(ComplexKeyGeneratorMixin,
                             LocalCacheBase,
                             AbstractExtendedLocalCache):
    """
    extended local cache base class.

    this type of caches are same as `LocalCacheBase` type but it also considers
    method inputs, current user and component key in key generation.
    """

    def __init__(self, *args, **options):
        """
        initializes an instance of ExtendedLocalCacheBase.

        :keyword bool consider_user: specifies that current user must also be
                                     included in cache key. if not provided, will
                                     be get from `caching` config store.

        :raises CacheNameIsRequiredError: cache name is required error.
        :raises InvalidCacheContainerTypeError: invalid cache container type error.
        :raises InvalidCacheItemTypeError: invalid cache item type error.
        """

        super().__init__(*args, **options)

        consider_user = options.get('consider_user')
        if consider_user is None:
            configs = self._get_configs()
            consider_user = configs['consider_user']

        self._consider_user = consider_user

    def _get_default_configs(self):
        """
        gets the defaults configs of this cache from `caching` config store.

        :rtype: dict
        """

        return config_services.get_section('caching', 'extended.permanent')

    def _try_set(self, value, func, inputs, kw_inputs, *args, **options):
        """
        sets a new value into cached items.

        this method will generate cache key from given type and function
        and inputs. it also considers current component key in key generation.

        :param object value: value to be cached.
        :param function func: function to cache its result.
        :param tuple inputs: function positional arguments.
        :param dict kw_inputs: function keyword arguments.

        :keyword bool consider_user: specifies that current user must be included in
                                     key generation. it will be get from `caching` config
                                     store if not provided.
        """

        key = self.generate_key(func, inputs, kw_inputs, *args, **options)
        self.set(key, value, **options)

    def _try_get(self, func, inputs, kw_inputs, *args, default=None, **options):
        """
        gets the value from cache.

        this method will generate cache key from given type and function
        and inputs. it also considers current component key in key generation.
        if key does not exist, it returns None or the specified default value.

        :param function func: function to to get its result.
        :param tuple inputs: function positional arguments.
        :param dict kw_inputs: function keyword arguments.
        :param object default: value to be returned if key is not present.

        :keyword bool consider_user: specifies that current user must be included in
                                     key generation. it will be get from `caching` config
                                     store if not provided.

        :returns: object
        """

        key = self.generate_key(func, inputs, kw_inputs, *args, **options)
        return self.get(key, default, **options)

    @property
    def consider_user(self):
        """
        gets the consider user value for this cache.

        :rtype: bool
        """

        return self._consider_user

    @property
    def stats(self):
        """
        get the statistic info about cached items.

        :returns: dict(int count: items count,
                       datetime last_cleared_time: last cleared time,
                       bool persistent: persistent cache,
                       bool consider_user: consider user in cache key)
        :rtype: dict
        """

        base_stats = super().stats
        stats = dict(consider_user=self.consider_user)
        base_stats.update(stats)
        return base_stats


class ComplexLocalCacheBase(ExtendedLocalCacheBase, AbstractComplexLocalCache):
    """
    complex local cache base class.

    this type of caches will also consider method inputs, current user and
    component key in key generation. this is useful for caching items that
    change during application runtime based on different inputs and variables.

    it also supports expire time and size limit for cached items.
    it also keeps a deep copy of the value in the cache.
    it also provides statistic info about hit or missed caches.
    it also supports persistent mode to save cached values into
    database on application shutdown and load them back on next startup.
    """

    # a lock type to be used on clearing cached items when cache limit is reached.
    clearance_lock_class = None

    # a lock type to be used on persisting or loading cached items to or from database.
    # this will only be used in persistent caches.
    persistent_lock_class = None

    def __init__(self, *args, **options):
        """
        initializes an instance of ComplexLocalCacheBase.

        :keyword int limit: limit count of cached items.
                            if not provided, it will be get
                            from `caching` config store.
                            if you want to remove count limit,
                            you could pass `caching.globals.NO_LIMIT`
                            as input.

        :keyword int expire: default expire time of cached items.
                             if not provided, it will be get
                             from `caching` config store.

        :keyword bool use_lifo: specifies that items of the cache must
                                be removed in lifo order. if not provided,
                                it will be get from `caching` config store.

        :keyword int clear_count: number of old items to be removed from cache when
                                  the cache is full. if not provided, it will be get
                                  from `caching` config store.
                                  note that reducing this value to extremely low values
                                  will cause a performance issue when the cache becomes full.

        :keyword bool consider_user: specifies that current user must also be
                                     included in cache key. if not provided, will
                                     be get from `caching` config store.

        :keyword bool persistent: specifies that cached items must be persisted to
                                  database on application shutdown, and loaded back
                                  on application startup. if not provided, will be
                                  get from `caching` config store.

        :keyword int chunk_size: chunk size to insert values for persistent caches.
                                 after each chunk, store will be flushed.
                                 if not provided, will be get from `caching` config store.

        :keyword bool refreshable: specifies that cached item's expire time must be
                                   extended on each hit. if not provided, will be get
                                   from `caching` config store.

        :raises CacheNameIsRequiredError: cache name is required error.
        :raises InvalidCacheContainerTypeError: invalid cache container type error.
        :raises InvalidCacheItemTypeError: invalid cache item type error.
        :raises CacheClearanceLockTypeIsRequiredError: cache clearance lock type
                                                       is required error.
        :raises InvalidCacheLimitError: invalid cache limit error.
        :raises InvalidCacheExpireTimeError: invalid cache expire time error.
        :raises InvalidCacheClearCountError: invalid cache clear count error.
        :raises InvalidChunkSizeError: invalid chunk size error.
        :raises CachePersistentLockTypeIsRequiredError: cache persistent lock
                                                        type is required error.
        """

        super().__init__(*args, **options)

        if self.cache_item_class is None or \
                not issubclass(self.cache_item_class, ComplexLocalCacheItemBase):
            raise InvalidCacheItemTypeError('Provided cache item [{item}] for cache '
                                            '[{name}] is not a subclass of [{base}].'
                                            .format(item=self.cache_item_class,
                                                    name=self.get_name(),
                                                    base=ComplexLocalCacheItemBase))

        if self.clearance_lock_class is None:
            raise CacheClearanceLockTypeIsRequiredError('Cache clearance lock type for '
                                                        'cache [{name}] is required.'
                                                        .format(name=self.get_name()))

        self._clearance_lock = self.clearance_lock_class()
        self._miss_count = 0
        self._hit_count = 0

        limit = options.get('limit')
        expire = options.get('expire')
        use_lifo = options.get('use_lifo')
        clear_count = options.get('clear_count')
        persistent = options.get('persistent')
        chunk_size = options.get('chunk_size')
        refreshable = options.get('refreshable')

        configs = self._get_configs()
        if limit is None:
            limit = configs['limit']

        if expire is None:
            expire = configs['expire']

        if use_lifo is None:
            use_lifo = configs['use_lifo']

        if clear_count is None:
            clear_count = configs['clear_count']

        if persistent is None:
            persistent = configs['persistent']

        if chunk_size is None:
            chunk_size = configs['chunk_size']

        if refreshable is None:
            refreshable = configs['refreshable']

        if limit != NO_LIMIT and limit <= 0:
            raise InvalidCacheLimitError('Cache limit for cache [{name}] '
                                         'must be a positive integer.'
                                         .format(name=self.get_name()))

        if expire <= 0:
            raise InvalidCacheExpireTimeError('Cache expire time for cache [{name}] '
                                              'must be a positive integer.'
                                              .format(name=self.get_name()))

        if clear_count <= 0:
            raise InvalidCacheClearCountError('Cache clear count for cache [{name}] '
                                              'must be a positive integer.'
                                              .format(name=self.get_name()))

        if persistent is True and chunk_size is not None and chunk_size <= 0:
            raise InvalidChunkSizeError('Persistent cache chunk size for cache '
                                        '[{name}] must be a positive integer'
                                        .format(name=self.get_name()))

        if persistent is True and self.persistent_lock_class is None:
            raise CachePersistentLockTypeIsRequiredError('Cache persistent lock type for '
                                                         'cache [{name}] is required.'
                                                         .format(name=self.get_name()))

        self._persistent_lock = self.persistent_lock_class()
        self._limit = limit
        self._expire = expire
        self._use_lifo = use_lifo
        self._clear_count = clear_count
        self._persistent = persistent
        self._chunk_size = chunk_size
        self._refreshable = refreshable

    def _get_default_configs(self):
        """
        gets the defaults configs of this cache from `caching` config store.

        :rtype: dict
        """

        return config_services.get_section('caching', 'complex')

    def _get_hit_ratio(self):
        """
        gets hit ratio for this cache in percentage.

        :rtype: float
        """

        hit = self.hit_count
        miss = self.miss_count
        if hit == 0 and miss == 0:
            return 0

        ratio = hit / (hit + miss)
        return ratio * 100

    def _is_cleaning_queue(self):
        """
        gets a value indicating that a thread is busy cleaning old cached items.

        :rtype: bool
        """

        return self._clearance_lock.locked() is True

    def _manage_queue(self):
        """
        manages this cache's queue and removes old items if required.

        the operation will be run in a new thread and does not block the current request.
        """

        if self._is_cleaning_queue() is True:
            return

        cleaner = Thread(target=self._remove_old_items)
        cleaner.start()

    def _remove_old_items(self):
        """
        removes old items from cache.

        the number of old items to be removed is calculated
        considering the `caching` config store.
        """

        if self._is_cleaning_queue() is True:
            return

        with self._clearance_lock:
            is_full, count = self.is_full
            if is_full is True:
                to_be_removed = self._container.slice(count, self._use_lifo)
                for key in to_be_removed:
                    self.remove(key)

    def _validate_persisting(self, version):
        """
        validates current cache for persisting.

        :param str version: version to be saved or loaded.

        :raises CacheIsNotPersistentError: cache is not persistent error.
        :raises CacheVersionIsRequiredError: cache version is required error.
        """

        if self._persistent is not True:
            raise CacheIsNotPersistentError('Cache [{name}] is not persistent.'
                                            .format(name=self.get_name()))

        if version in (None, '') or version.isspace():
            raise CacheVersionIsRequiredError('Cache version is required for persisting '
                                              'cache [{name}] into database.'
                                              .format(name=self.get_name()))

    def _delete_loaded_caches(self, version, shard_name, **options):
        """
        deletes loaded caches from database.

        :param str version: version to remove its caches.
        :param str shard_name: shard name to remove its caches.
        """

        store = get_current_store()
        store.query(CacheItemEntity).filter_by(cache_name=self.get_name(),
                                               version=version,
                                               shard_name=shard_name).delete()

    def set(self, key, value, **options):
        """
        sets a new value into cached items.

        :param object key: hashable key of the cache to be registered.
        :param object value: value to be cached.

        :keyword int expire: expire time for given key in milliseconds.
                             defaults to `expire` attribute if not provided.

        :keyword bool refreshable: specifies that cached item's expire time must be
                                   extended on each hit. defaults to `refreshable`
                                   attribute if not provided.
        """

        expire = options.get('expire')
        if expire is None:
            options.update(expire=self.expire)

        refreshable = options.get('refreshable')
        if refreshable is None:
            options.update(refreshable=self.refreshable)

        is_full, count = self.is_full
        if is_full is True:
            self._manage_queue()

        super().set(key, value, **options)

    def get(self, key, default=None, **options):
        """
        gets the value from cache.

        if key does not exist, it returns None or the specified default value.

        :param object key: hashable key to get its value from cache.
        :param object default: value to be returned if key is not present.

        :returns: object
        """

        key_hash = hash(key)
        result = self._container.get(key_hash, default)
        if result in (None, default):
            self._miss_count = self._miss_count + 1
            return result

        if result.is_expired is True:
            self._miss_count = self._miss_count + 1
            self.remove(key_hash)
            return default

        self._hit_count = self._hit_count + 1
        self._container.move_to_end(key_hash, not self._use_lifo)
        return result.value

    def persist(self, version, **options):
        """
        saves cached items of this cache into database.

        :param str version: version to be saved with cached items in database.

        :keyword bool clear: clear all caches after persisting them into database.
                             defaults to False if not provided.

        :raises CacheIsNotPersistentError: cache is not persistent error.
        :raises CacheVersionIsRequiredError: cache version is required error.
        """

        self._validate_persisting(version)

        with self._persistent_lock:
            shard_name = config_services.get('caching', 'general', 'shard_name')
            caches = []
            for item in self.values():
                if item.is_expired is False:
                    try:
                        pickled_item = pickle.dumps(item)
                        entity = CacheItemEntity()
                        entity.cache_name = self.get_name()
                        entity.shard_name = shard_name
                        entity.version = version
                        entity.key = item.key
                        entity.item = pickled_item
                        caches.append(entity)

                    except Exception as error:
                        self.LOGGER.exception(str(error))

            size = len(caches)
            if size > 0:
                bulk_services.insert(*caches, readable=SECURE_FALSE,
                                     chunk_size=self.chunk_size)

                self.LOGGER.debug('Cache [{name}] persisted into database. '
                                  'including [{count}] items.'
                                  .format(name=self.get_name(), count=size))

            clear = options.get('clear', False)
            if clear is True:
                self.clear()

    def load(self, version, **options):
        """
        loads cached items of this cache from database.

        :param str version: version of cached items to be loaded from database.

        :raises CacheIsNotPersistentError: cache is not persistent error.
        :raises CacheVersionIsRequiredError: cache version is required error.
        """

        self._validate_persisting(version)

        with self._persistent_lock:
            shard_name = config_services.get('caching', 'general', 'shard_name')
            store = get_current_store()
            expression = store.query(CacheItemEntity).filter_by(
                cache_name=self.get_name(),
                shard_name=shard_name,
                version=version).order_by(
                CacheItemEntity.created_on.desc())

            if self.limit != NO_LIMIT:
                expression = expression.limit(self.limit)

            items = expression.all()
            for entity in items:
                try:
                    cache_item = pickle.loads(entity.item)
                    self.set(cache_item.key, cache_item.value, expire=cache_item.expire)

                except Exception as error:
                    self.LOGGER.exception(str(error))

            size = len(items)
            if size > 0:
                self.LOGGER.debug('Cache [{name}] loaded from database. '
                                  'including [{count}] items.'
                                  .format(name=self.get_name(), count=size))

            self._delete_loaded_caches(version, shard_name)

    @property
    def is_full(self):
        """
        gets a value indicating the cache is full.

        it returns a tuple, first item is a boolean indicating the fullness of cache.
        the second item is the number of excess items in the cache that must be removed.
        the excess value will be calculated using `clear_count` property of this cache.

        :rtype: tuple[bool, int]
        """

        if self.limit == NO_LIMIT:
            return False

        count = self.count
        is_full = False
        excess = 0
        if count >= self.limit:
            is_full = True
            excess = count - self.limit + self._clear_count + 1

        return is_full, excess

    @property
    def expire(self):
        """
        gets default expire time value for this cache's items in milliseconds.

        :rtype: int
        """

        return self._expire

    @property
    def limit(self):
        """
        gets the size limit of this cache.

        :rtype: int
        """

        return self._limit

    @property
    def hit_count(self):
        """
        gets the hit count for this cache.

        :rtype: int
        """

        return self._hit_count

    @property
    def miss_count(self):
        """
        gets the miss count for this cache.

        :rtype: int
        """

        return self._miss_count

    @property
    def use_lifo(self):
        """
        gets the use lifo for this cache.

        :rtype: bool
        """

        return self._use_lifo

    @property
    def clear_count(self):
        """
        gets the clear count for this cache.

        :rtype: int
        """

        return self._clear_count

    @property
    def refreshable(self):
        """
        gets a value indicating that cached item's expire time will be extended on each hit.

        :rtype: bool
        """

        return self._refreshable

    @property
    def stats(self):
        """
        get the statistic info about cached items.

        :returns: dict(int count: items count,
                       datetime last_cleared_time: last cleared time,
                       bool persistent: persistent cache,
                       bool consider_user: consider user in cache key,
                       int hit: hit count,
                       int miss: miss count,
                       str hit_ratio: hit ratio,
                       int limit: items count limit,
                       int expire: items default expire time,
                       bool refreshable: items default refreshable value.
                       bool use_lifo: use lifo order,
                       int clear_count: clear count,
                       int chunk_size: chunk size)
        :rtype: dict
        """

        base_stats = super().stats
        hit_ratio = self._get_hit_ratio()
        hit_ratio = '{:0.1f}%'.format(hit_ratio)
        stats = dict(hit=self.hit_count,
                     miss=self.miss_count,
                     hit_ratio=hit_ratio,
                     limit=self.limit,
                     expire=self.expire,
                     refreshable=self.refreshable,
                     use_lifo=self.use_lifo,
                     clear_count=self.clear_count,
                     chunk_size=self.chunk_size)

        base_stats.update(stats)
        return base_stats

    @property
    def persistent(self):
        """
        gets a value indicating that cached items must be persisted to database on shutdown.

        :rtype: bool
        """

        return self._persistent

    @property
    def chunk_size(self):
        """
        gets the chunk size for this cache.

        :rtype: int
        """

        return self._chunk_size



# -*- coding: utf-8 -*-
"""
template manager module.
"""

from pyrin.template import TemplatePackage
from pyrin.template.interface import AbstractTemplateHandler
from pyrin.utils.custom_print import print_warning
from pyrin.core.structs import Manager, DTO
from pyrin.template.exceptions import InvalidTemplateHandlerTypeError, \
    DuplicatedTemplateHandlerError, TemplateHandlerNotFoundError


class TemplateManager(Manager):
    """
    template manager class.
    """

    package_class = TemplatePackage

    def __init__(self):
        """
        initializes an instance of TemplateManager.
        """

        super().__init__()

        # a dictionary containing template handlers.
        # in the form of: {str handler_name: AbstractTemplateHandler handler}
        self._template_handlers = DTO()

    def register_template_handler(self, instance, **options):
        """
        registers a new template handler or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding a template handler which is already registered.

        :param AbstractTemplateHandler instance: template handler to be registered.
                                                 it must be an instance of
                                                 AbstractTemplateHandler.

        :keyword bool replace: specifies that if there is another registered
                               template handler with the same name, replace it
                               with the new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidTemplateHandlerTypeError: invalid template handler type error.
        :raises DuplicatedTemplateHandlerError: duplicated template handler error.
        """

        if not isinstance(instance, AbstractTemplateHandler):
            raise InvalidTemplateHandlerTypeError('Input parameter [{instance}] is '
                                                  'not an instance of [{handler}].'
                                                  .format(instance=instance,
                                                          handler=AbstractTemplateHandler))

        if instance.name in self._template_handlers:
            old_instance = self._template_handlers.get(instance.name)
            replace = options.get('replace', False)
            if replace is not True:
                raise DuplicatedTemplateHandlerError('There is another registered '
                                                     'template handler with name [{name}] '
                                                     'but "replace" option is not set, so '
                                                     'template handler [{instance}] could '
                                                     'not be registered.'
                                                     .format(name=instance.name,
                                                             instance=instance))

            print_warning('Template handler [{old_instance}] is going '
                          'to be replaced by [{new_instance}].'
                          .format(old_instance=old_instance,
                                  new_instance=instance))

        self._template_handlers[instance.name] = instance

    def _get_template_handler(self, name):
        """
        gets a template handler with the given name.

        if not available, it raises an error.

        :param str name: template handler name to get its instance.

        :raises TemplateHandlerNotFoundError: template handler not found error.

        :rtype: AbstractTemplateHandler
        """

        if name not in self._template_handlers:
            raise TemplateHandlerNotFoundError('Template handler [{name}] not found.'
                                               .format(name=name))

        return self._template_handlers[name]

    def create(self, handler_name, *args, **kwargs):
        """
        creates the template using the given template handler name.

        :param str handler_name: handler name to be used.
        :param object args: arguments that should be passed to template handler.

        :keyword object kwargs: keyword arguments that should be passed to template handler.

        :raises TemplateHandlerNotFoundError: template handler not found error.
        """

        handler = self._get_template_handler(handler_name)
        return handler.create(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
session manager module.
"""

from flask import request
from flask.ctx import has_request_context

from pyrin.core.structs import DTO, Manager
from pyrin.security.session import SessionPackage
from pyrin.security.session.exceptions import InvalidUserError, \
    CouldNotOverwriteCurrentUserError, InvalidComponentCustomKeyError


class SessionManager(Manager):
    """
    session manager class.
    """

    package_class = SessionPackage

    def get_current_user(self):
        """
        gets current user.
        """

        return self.get_current_request().user

    def set_current_user(self, user):
        """
        sets current user.

        :param user: user object.

        :raises InvalidUserError: invalid user error.
        :raises CouldNotOverwriteCurrentUserError: could not overwrite current user error.
        """

        if user is None:
            raise InvalidUserError('Request user could not be None.')

        if self.get_current_user() is not None:
            raise CouldNotOverwriteCurrentUserError('User has been already set for '
                                                    'current request, it could not '
                                                    'be overwritten.')

        self.get_current_request().user = user

    def get_current_request(self):
        """
        gets current request object.

        :raises RuntimeError: runtime error.

        :rtype: CoreRequest
        """

        return request

    def get_current_request_id(self):
        """
        gets current request id.

        this method is implemented to overcome the hash problem of requests.
        `CoreRequest` objects are hashable themselves, but when they used as a dict key
        some hash collisions will occur. so we have to expose the exact request id to
        be able to use it as a dict key in places such as database scoped sessions.

        :raises RuntimeError: runtime error.

        :rtype: uuid.UUID
        """

        return self.get_current_request().request_id

    def add_request_context(self, key, value, **options):
        """
        adds the given key/value pair into current request context.

        :param str key: key name to be added.
        :param object value: value to be added.

        :keyword bool replace: specifies that if a key with the same name
                               is already present, replace it. otherwise
                               raise an error. defaults to False if not provided.

        :raises InvalidRequestContextKeyNameError: invalid request context key name error.
        :raises RequestContextKeyIsAlreadyPresentError: request context key is
                                                        already present error.
        """

        self.get_current_request().add_context(key, value, **options)

    def get_request_context(self, key, default=None):
        """
        gets the value for given key from current request context.

        it gets the default value if key is not present in the request context.

        :param str key: key name to get its value.
        :param object default: a value to be returned if the provided
                               key is not present in request context.

        :returns: object
        """

        return self.get_current_request().get_context(key, default)

    def remove_request_context(self, key):
        """
        removes the specified key from current request context if available.

        :param str key: key name to be removed from request context.
        """

        self.get_current_request().remove_context(key)

    def is_fresh(self):
        """
        gets a value indicating that current request has a fresh authentication.

        fresh authentication means an authentication which is done by providing
        user credentials to server.

        :rtype: bool
        """

        return self.get_current_token_payload().get('is_fresh', False)

    def get_current_token_payload(self):
        """
        gets current request's token payload.

        :rtype: dict
        """

        return self.get_request_context('token_payload', DTO())

    def get_current_token_header(self):
        """
        gets current request's token header.

        :rtype: dict
        """

        return self.get_request_context('token_header', DTO())

    def set_component_custom_key(self, value):
        """
        sets the component custom key.

        :param object value: component custom key value.

        :raises InvalidComponentCustomKeyError: invalid component custom key error.
        """

        if value is None:
            raise InvalidComponentCustomKeyError('Component custom key could not be None.')

        self.get_current_request().component_custom_key = value

    def get_component_custom_key(self):
        """
        gets component custom key.

        :rtype: object
        """

        return self.get_current_request().component_custom_key

    def get_safe_component_custom_key(self):
        """
        gets component custom key in a safe manner.

        meaning that if the request does not exist in current context, it will
        return a None object instead of raising an error.

        :rtype: object
        """

        current_request = self.get_safe_current_request()
        if current_request is None:
            return None

        return current_request.component_custom_key

    def get_safe_current_request(self):
        """
        gets current request object in a safe manner.

        meaning that if the request does not exist in current context, it will
        return a None object instead of raising an error.

        :rtype: CoreRequest
        """

        if self.is_request_context_available() is True:
            return self.get_current_request()

        return None

    def get_safe_current_user(self):
        """
        gets current user in a safe manner.

        meaning that if the request does not exist in current context, it will
        return a None object instead of raising an error.
        """

        current_request = self.get_safe_current_request()
        if current_request is None:
            return None

        return current_request.user

    def get_safe_cacheable_current_user(self):
        """
        gets cacheable current user in a safe manner.

        meaning that if the request does not exist in current context, it will
        return a None object instead of raising an error.
        """

        current_request = self.get_safe_current_request()
        if current_request is None:
            return None

        return current_request.cacheable_user

    def is_request_context_available(self):
        """
        gets a value indicating that request context is available for usage.

        :rtype: bool
        """

        return has_request_context()
